##############################################################################
#
# Copyright (c) 2001, 2002, 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""The StorageServer class and the exception that it may raise.

This server acts as a front-end for one or more real storages, like
file storage or Berkeley storage.

TODO:  Need some basic access control-- a declaration of the methods
exported for invocation by the server.
"""
import codecs
import itertools
import logging
import os
import sys
import tempfile
import threading
import time
import warnings

import ZODB.blob
import ZODB.event
import ZODB.serialize
import ZODB.TimeStamp
import zope.interface
from ZODB.Connection import TransactionMetaData
from ZODB.loglevels import BLATHER
from ZODB.POSException import ConflictError
from ZODB.POSException import ReadOnlyError
from ZODB.POSException import StorageError
from ZODB.POSException import StorageTransactionError
from ZODB.POSException import TransactionError
from ZODB.serialize import referencesf
from ZODB.utils import Lock
from ZODB.utils import RLock
from ZODB.utils import p64
from ZODB.utils import u64
from ZODB.utils import z64
from zodbpickle.pickle import Pickler

from ZEO._compat import Unpickler
from ZEO.asyncio.server import Acceptor
from ZEO.asyncio.server import Delay
from ZEO.asyncio.server import MTDelay
from ZEO.asyncio.server import Result
from ZEO.monitor import StorageStats


# multi-threaded acceptor was opt-in option, but later was deprecated and
# removed. Warn users that try to activate multi-threaded server mode to
# explicitly let them know they no longer get it and why.
if os.environ.get("ZEO_MTACCEPTOR"):
    warnings.warn('The mtacceptor module is no longer supported because it '
                  'was subject to data corruption bugs. $ZEO_MTACCEPTOR no '
                  'longer has any effect. Please see '
                  'github.com/zopefoundation/ZEO/issues/209 for details.',
                  DeprecationWarning)

logger = logging.getLogger('ZEO.StorageServer')


def log(message, level=logging.INFO, label='', exc_info=False):
    """Internal helper to log a message."""
    if label:
        message = f'({label}) {message}'
    logger.log(level, message, exc_info=exc_info)


class StorageServerError(StorageError):
    """Error reported when an unpicklable exception is raised."""


registered_methods = {
    'get_info', 'lastTransaction',
     'getInvalidations', 'new_oids', 'pack', 'loadBefore', 'storea',
     'checkCurrentSerialInTransaction', 'restorea', 'storeBlobStart',
     'storeBlobChunk', 'storeBlobEnd', 'storeBlobShared',
     'deleteObject', 'tpc_begin', 'vote', 'tpc_finish', 'tpc_abort',
     'history', 'record_iternext', 'sendBlob', 'getTid', 'loadSerial',
     'new_oid', 'undoa', 'undoLog', 'undoInfo', 'iterator_start',
     'iterator_next', 'iterator_record_start', 'iterator_record_next',
     'iterator_gc', 'server_status', 'set_client_label', 'ping'}


class ZEOStorage:
    """Proxy to underlying storage for a single remote client."""

    connected = connection = stats = storage = storage_id = transaction = None
    blob_tempfile = None
    log_label = 'unconnected'
    locked = False             # Don't have storage lock
    verifying = 0

    def __init__(self, server, read_only=0):
        self.server = server
        self.client_conflict_resolution = server.client_conflict_resolution
        # timeout and stats will be initialized in register()
        self.read_only = read_only
        self._iterators = {}
        self._iterator_ids = itertools.count()
        # Stores the last item that was handed out for a
        # transaction iterator.
        self._txn_iterators_last = {}

    def set_database(self, database):
        self.database = database

    def notify_connected(self, conn):
        self.connection = conn
        self.call_soon_threadsafe = conn.call_soon_threadsafe
        self.connected = True
        assert conn.protocol_version is not None
        self.log_label = _addr_label(conn.addr)
        self.async_ = conn.async_
        self.async_threadsafe = conn.async_threadsafe

    def notify_disconnected(self):
        # When this storage closes, we must ensure that it aborts
        # any pending transaction.
        if self.transaction is not None:
            self.log("disconnected during %s transaction"
                     % (self.locked and 'locked' or 'unlocked'))
            self.tpc_abort(self.transaction.id)
        else:
            self.log("disconnected")

        self.connected = False
        self.server.close_conn(self)

    def __repr__(self):
        tid = self.transaction and repr(self.transaction.id)
        if self.storage:
            stid = (self.tpc_transaction() and
                    repr(self.tpc_transaction().id))
        else:
            stid = None
        name = self.__class__.__name__
        return f'<{name} {id(self):X} trans={tid} s_trans={stid}>'

    def log(self, msg, level=logging.INFO, exc_info=False):
        log(msg, level=level, label=self.log_label, exc_info=exc_info)

    def setup_delegation(self):
        """Delegate several methods to the storage
        """
        # Called from register

        storage = self.storage

        info = self.get_info()

        if not info['supportsUndo']:
            self.undoLog = self.undoInfo = lambda *a, **k: ()

        # XXX deprecated: but ZODB tests use getTid. They shouldn't
        self.getTid = storage.getTid

        self.loadSerial = storage.loadSerial
        record_iternext = getattr(storage, 'record_iternext', None)
        if record_iternext is not None:
            self.record_iternext = record_iternext
        self.lastTransaction = storage.lastTransaction

        try:
            self.tpc_transaction = storage.tpc_transaction
        except AttributeError:
            if hasattr(storage, '_transaction'):
                log("Storage %r doesn't have a tpc_transaction method.\n"
                    "See ZEO.interfaces.IServeable."
                    "Falling back to using _transaction attribute, which\n."
                    "is icky.",
                    logging.ERROR)
                self.tpc_transaction = lambda: storage._transaction
            else:
                raise

        self.connection.methods = registered_methods

    def history(self, tid, size=1):
        # This caters for storages which still accept
        # a version parameter.
        return self.storage.history(tid, size=size)

    def _check_tid(self, tid, exc=None):
        if self.read_only:
            raise ReadOnlyError()
        if self.transaction is None:
            caller = sys._getframe().f_back.f_code.co_name
            # ``tpc_abort`` is allowed to be called with invalid transaction
            # ``vote`` relies on this
            if caller != "tpc_abort":
                self.log("no current transaction: %s()" % caller,
                         level=logging.WARNING)
            if exc is not None:
                raise exc(None, tid)
            else:
                return 0
        if self.transaction.id != tid:
            caller = sys._getframe().f_back.f_code.co_name
            # ``tpc_abort`` is allowed to be called with invalid transaction
            if caller != "tpc_abort":
                self.log("%s(%s) invalid; current transaction = %s" %
                         (caller, repr(tid), repr(self.transaction.id)),
                         logging.WARNING)
            if exc is not None:
                raise exc(self.transaction.id, tid)
            else:
                return 0
        return 1

    def register(self, storage_id, read_only):
        """Select the storage that this client will use

        This method must be the first one called by the client.
        For authenticated storages this method will be called by the client
        immediately after authentication is finished.
        """
        if self.storage is not None:
            self.log("duplicate register() call")
            raise ValueError("duplicate register() call")

        storage = self.server.storages.get(storage_id)
        if storage is None:
            self.log("unknown storage_id: %s" % storage_id)
            raise ValueError("unknown storage: %s" % storage_id)

        if not read_only and (self.read_only or storage.isReadOnly()):
            raise ReadOnlyError()

        self.read_only = self.read_only or read_only
        self.storage_id = storage_id
        self.storage = storage
        self.setup_delegation()
        self.stats = self.server.register_connection(storage_id, self)
        self.lock_manager = self.server.lock_managers[storage_id]

        return self.lastTransaction()

    def get_info(self):
        storage = self.storage

        supportsUndo = (getattr(storage, 'supportsUndo', lambda: False)()
                        and self.connection.protocol_version[1:] >= b'310')

        # Communicate the backend storage interfaces to the client
        storage_provides = zope.interface.providedBy(storage)
        interfaces = []
        for candidate in storage_provides.__iro__:
            interfaces.append((candidate.__module__, candidate.__name__))

        return {'length': len(storage),
                'size': storage.getSize(),
                'name': storage.getName(),
                'supportsUndo': supportsUndo,
                'supports_record_iternext': hasattr(self, 'record_iternext'),
                'interfaces': tuple(interfaces),
                }

    def get_size_info(self):
        return {'length': len(self.storage),
                'size': self.storage.getSize(),
                }

    def loadBefore(self, oid, tid):
        self.stats.loads += 1
        return self.storage.loadBefore(oid, tid)

    def getInvalidations(self, tid):
        invtid, invlist = self.server.get_invalidations(self.storage_id, tid)
        if invtid is None:
            return None
        self.log("Return %d invalidations up to tid %s"
                 % (len(invlist), u64(invtid)))
        return invtid, invlist

    def pack(self, time, wait=1):
        # Yes, you can pack a read-only server or storage!
        if wait:
            return run_in_thread(self._pack_impl, time)
        else:
            # If the client isn't waiting for a reply, start a thread
            # and forget about it.
            t = threading.Thread(target=self._pack_impl, args=(time,))
            t.name = "zeo storage packing thread"
            t.start()
            return None

    def _pack_impl(self, time):
        self.log("pack(time=%s) started..." % repr(time))
        self.storage.pack(time, referencesf)
        self.log("pack(time=%s) complete" % repr(time))
        # Broadcast new size statistics
        self.server.broadcast_info(self.storage_id, self.get_size_info())

    def new_oids(self, n=100):
        """Return a sequence of n new oids, where n defaults to 100"""
        n = min(n, 100)
        if self.read_only:
            raise ReadOnlyError()
        if n <= 0:
            n = 1
        return [self.storage.new_oid() for i in range(n)]

    # undoLog and undoInfo are potentially slow methods

    def undoInfo(self, first, last, spec):
        return run_in_thread(self.storage.undoInfo, first, last, spec)

    def undoLog(self, first, last):
        return run_in_thread(self.storage.undoLog, first, last)

    def tpc_begin(self, id, user, description, ext, tid=None, status=" "):
        if self.read_only:
            raise ReadOnlyError()
        if self.transaction is not None:
            if self.transaction.id == id:
                self.log("duplicate tpc_begin(%s)" % repr(id))
                return
            else:
                raise StorageTransactionError("Multiple simultaneous tpc_begin"
                                              " requests from one client.")

        t = TransactionMetaData(user, description, ext)
        t.id = id

        self.serials = []
        self.conflicts = {}
        self.invalidated = []
        self.txnlog = CommitLog()
        self.blob_log = []
        self.tid = tid
        self.status = status
        self.stats.active_txns += 1

        # Assign the transaction attribute last. This is so we don't
        # think we've entered TPC until everything is set.  Why?
        # Because if we have an error after this, the server will
        # think it is in TPC and the client will think it isn't.  At
        # that point, the client will keep trying to enter TPC and
        # server won't let it.  Errors *after* the tpc_begin call will
        # cause the client to abort the transaction.
        # (Also see https://bugs.launchpad.net/zodb/+bug/374737.)
        self.transaction = t

    def tpc_finish(self, id):
        if not self._check_tid(id):
            return
        assert self.locked, "finished called wo lock"

        self.stats.commits += 1
        self.storage.tpc_finish(self.transaction, self._invalidate)
        self.async_('info', self.get_size_info())
        # Note that the tid is still current because we still hold the
        # commit lock. We'll relinquish it in _clear_transaction.
        tid = self.storage.lastTransaction()
        # Return the tid, for cache invalidation optimization
        return Result(tid, self._clear_transaction)

    def _invalidate(self, tid):
        self.server.invalidate(self, self.storage_id, tid, self.invalidated)

    def tpc_abort(self, tid):
        if not self._check_tid(tid):
            return
        self.stats.aborts += 1
        self.storage.tpc_abort(self.transaction)
        self._clear_transaction()

    def _clear_transaction(self):
        # Common code at end of tpc_finish() and tpc_abort()
        self.lock_manager.release(self)
        self.transaction = None
        self.stats.active_txns -= 1
        if self.txnlog is not None:
            self.txnlog.close()
            self.txnlog = None
            for oid, oldserial, data, blobfilename in self.blob_log:
                ZODB.blob.remove_committed(blobfilename)
            del self.blob_log

    def vote(self, tid):
        self._check_tid(tid, exc=StorageTransactionError)
        return self.lock_manager.lock(self, self._vote)

    def _vote(self, delay=None):
        # Called from client thread

        if not self.connected:
            return  # We're disconnected

        try:
            self.log(
                "Preparing to commit transaction: %d objects, %d bytes"
                % (self.txnlog.stores, self.txnlog.size()),
                level=BLATHER)

            if (self.tid is not None) or (self.status != ' '):
                self.storage.tpc_begin(self.transaction,
                                       self.tid, self.status)
            else:
                self.storage.tpc_begin(self.transaction)

            for op, args in self.txnlog:
                getattr(self, op)(*args)

            # Blob support
            while self.blob_log:
                oid, oldserial, data, blobfilename = self.blob_log.pop()
                self._store(oid, oldserial, data, blobfilename)

            if not self.conflicts:
                try:
                    serials = self.storage.tpc_vote(self.transaction)
                except ConflictError as err:
                    if self.client_conflict_resolution and \
                       err.oid and err.serials and err.data:
                        self.conflicts[err.oid] = dict(
                            oid=err.oid, serials=err.serials, data=err.data)
                    else:
                        raise
                else:
                    if serials:
                        self.serials.extend(serials)

            if self.conflicts:
                self.storage.tpc_abort(self.transaction)
                return list(self.conflicts.values())
            else:
                self.locked = True  # signal to lock manager to hold lock
                return self.serials

        except Exception as err:
            self.storage.tpc_abort(self.transaction)
            self._clear_transaction()

            if isinstance(err, ConflictError):
                self.stats.conflicts += 1
                self.log("conflict error %s" % err, BLATHER)

            if not isinstance(err, TransactionError):
                logger.exception("While voting")

            raise

    # The public methods of the ZEO client API do not do the real work.
    # They defer work until after the storage lock has been acquired.
    # Most of the real implementations are in methods beginning with
    # an _.

    def deleteObject(self, oid, serial, id):
        self._check_tid(id, exc=StorageTransactionError)
        self.stats.stores += 1
        self.txnlog.delete(oid, serial)

    def storea(self, oid, serial, data, id):
        self._check_tid(id, exc=StorageTransactionError)
        self.stats.stores += 1
        self.txnlog.store(oid, serial, data)

    def checkCurrentSerialInTransaction(self, oid, serial, id):
        self._check_tid(id, exc=StorageTransactionError)
        self.txnlog.checkread(oid, serial)

    def restorea(self, oid, serial, data, prev_txn, id):
        self._check_tid(id, exc=StorageTransactionError)
        self.stats.stores += 1
        self.txnlog.restore(oid, serial, data, prev_txn)

    def storeBlobStart(self):
        assert self.blob_tempfile is None
        self.blob_tempfile = tempfile.mkstemp(
            dir=self.storage.temporaryDirectory())

    def storeBlobChunk(self, chunk):
        os.write(self.blob_tempfile[0], chunk)

    def storeBlobEnd(self, oid, serial, data, id):
        self._check_tid(id, exc=StorageTransactionError)
        assert self.txnlog is not None  # effectively not allowed after undo
        fd, tempname = self.blob_tempfile
        self.blob_tempfile = None
        os.close(fd)
        self.blob_log.append((oid, serial, data, tempname))

    def storeBlobShared(self, oid, serial, data, filename, id):
        self._check_tid(id, exc=StorageTransactionError)
        assert self.txnlog is not None  # effectively not allowed after undo

        # Reconstruct the full path from the filename in the OID directory
        if os.path.sep in filename or \
           not (filename.endswith('.tmp') or filename[:-1].endswith('.tmp')):
            logger.critical(
                "We're under attack! (bad filename to storeBlobShared, %r)",
                filename)
            raise ValueError(filename)

        filename = os.path.join(self.storage.fshelper.getPathForOID(oid),
                                filename)
        self.blob_log.append((oid, serial, data, filename))

    def sendBlob(self, oid, serial):
        logger.debug("Blob requested %r, %r", oid, serial)
        blobfilename = self.storage.loadBlob(oid, serial)
        logger.debug("Blob ready %r, %r", oid, serial)

        def store():
            logger.debug("send `receiveBlobStart` for %r, %r", oid, serial)
            yield ('receiveBlobStart', (oid, serial))
            with open(blobfilename, 'rb') as f:
                while 1:
                    chunk = f.read(59000)
                    if not chunk:
                        break
                    logger.debug("send `receiveBlobChunk` for %r, %r",
                                 oid, serial)
                    yield ('receiveBlobChunk', (oid, serial, chunk, ))
            logger.debug("send `receiveBlobStop` for %r, %r", oid, serial)
            yield ('receiveBlobStop', (oid, serial))

        self.connection.call_async_iter(store())

    def undo(*a, **k):
        raise NotImplementedError

    def undoa(self, trans_id, tid):
        self._check_tid(tid, exc=StorageTransactionError)
        self.txnlog.undo(trans_id)

    def _delete(self, oid, serial):
        self.storage.deleteObject(oid, serial, self.transaction)

    def _checkread(self, oid, serial):
        self.storage.checkCurrentSerialInTransaction(
            oid, serial, self.transaction)

    def _store(self, oid, serial, data, blobfile=None):
        try:
            if blobfile is None:
                self.storage.store(oid, serial, data, '', self.transaction)
            else:
                self.storage.storeBlob(
                    oid, serial, data, blobfile, '', self.transaction)
        except ConflictError as err:
            if self.client_conflict_resolution and err.serials:
                self.conflicts[oid] = dict(
                    oid=oid, serials=err.serials, data=data)
            else:
                raise
        else:
            if oid in self.conflicts:
                del self.conflicts[oid]

            self.invalidated.append(oid)

    def _restore(self, oid, serial, data, prev_txn):
        self.storage.restore(oid, serial, data, '', prev_txn,
                             self.transaction)

    def _undo(self, trans_id):
        tid, oids = self.storage.undo(trans_id, self.transaction)
        self.invalidated.extend(oids)
        self.serials.extend(oids)

    # IStorageIteration support

    def iterator_start(self, start, stop):
        iid = next(self._iterator_ids)
        self._iterators[iid] = iter(self.storage.iterator(start, stop))
        return iid

    def iterator_next(self, iid):
        iterator = self._iterators[iid]
        try:
            info = next(iterator)
        except StopIteration:
            del self._iterators[iid]
            item = None
            if iid in self._txn_iterators_last:
                del self._txn_iterators_last[iid]
        else:
            item = (info.tid,
                    info.status,
                    info.user,
                    info.description,
                    info.extension)
            # Keep a reference to the last iterator result to allow starting a
            # record iterator off it.
            self._txn_iterators_last[iid] = info
        return item

    def iterator_record_start(self, txn_iid, tid):
        record_iid = next(self._iterator_ids)
        txn_info = self._txn_iterators_last[txn_iid]
        if txn_info.tid != tid:
            raise Exception(
                'Out-of-order request for record iterator for transaction %r'
                % tid)
        self._iterators[record_iid] = iter(txn_info)
        return record_iid

    def iterator_record_next(self, iid):
        iterator = self._iterators[iid]
        try:
            info = next(iterator)
        except StopIteration:
            del self._iterators[iid]
            item = None
        else:
            item = (info.oid,
                    info.tid,
                    info.data,
                    info.data_txn)
        return item

    def iterator_gc(self, iids):
        for iid in iids:
            it = self._iterators.pop(iid, None)
            if hasattr(it, "close"):
                it.close()

    def server_status(self):
        return self.server.server_status(self.storage_id)

    def set_client_label(self, label):
        self.log_label = str(label)+' '+_addr_label(self.connection.addr)

    def ruok(self):
        return self.server.ruok()

    def ping(self):
        pass


class StorageServerDB:
    """Adapts (StorageServer, storage_id) to ZODB.interfaces.IStorageWrapper.

    The class is used as ``DB`` emulation in a ``registerDB`` call;
    it allows the storage server to keep its cached data about a storage
    up to date and to keep the respective connections informed.

    In particular, the class is used in a ZEO fan-out situation,
    where a storage server calls registerDB on a ClientStorage.
    Note that in this case the methods are called from the Client-storage's
    IO thread, a separate thread from the storge-server connections.
    Thus, the methods need to be thread safe.
    """

    def __init__(self, server, storage_id):
        self.server = server
        self.storage_id = storage_id
        self.references = ZODB.serialize.referencesf

    def invalidate(self, tid, oids, version=''):
        if version:
            raise StorageServerError("Versions aren't supported.")
        storage_id = self.storage_id
        self.server.invalidate(None, storage_id, tid, oids)

    def invalidateCache(self):
        self.server._invalidateCache(self.storage_id)

    transform_record_data = untransform_record_data = lambda self, data: data


class StorageServer:

    """The server side implementation of ZEO.

    The StorageServer is the 'manager' for incoming connections.  Each
    connection is associated with its own ZEOStorage instance (defined
    below).  The StorageServer may handle multiple storages; each
    ZEOStorage instance only handles a single storage.
    """

    def __init__(self, addr, storages,
                 read_only=0,
                 invalidation_queue_size=100,
                 invalidation_age=None,
                 transaction_timeout=None,
                 ssl=None,
                 client_conflict_resolution=False,
                 Acceptor=Acceptor,
                 msgpack=False,
                 ):
        """StorageServer constructor.

        This is typically invoked from the start.py script.

        Arguments (the first two are required and positional):

        addr -- the address at which the server should listen.  This
            can be a tuple (host, port) to signify a TCP/IP connection
            or a pathname string to signify a Unix domain socket
            connection.  A hostname may be a DNS name or a dotted IP
            address.

        storages -- a dictionary giving the storage(s) to handle.  The
            keys are the storage names, the values are the storage
            instances, typically FileStorage or Berkeley storage
            instances.  By convention, storage names are typically
            strings representing small integers starting at '1'.

        read_only -- an optional flag saying whether the server should
            operate in read-only mode.  Defaults to false.  Note that
            even if the server is operating in writable mode,
            individual storages may still be read-only.  But if the
            server is in read-only mode, no write operations are
            allowed, even if the storages are writable.  Note that
            pack() is considered a read-only operation.

        invalidation_queue_size -- The storage server keeps a queue
            of the objects modified by the last N transactions, where
            N == invalidation_queue_size.  This queue is used to
            speed client cache verification when a client disconnects
            for a short period of time.

        invalidation_age --
            If the invalidation queue isn't big enough to support a
            quick verification, but the last transaction seen by a
            client is younger than the invalidation age, then
            invalidations will be computed by iterating over
            transactions later than the given transaction.

        transaction_timeout -- The maximum amount of time to wait for
            a transaction to commit after acquiring the storage lock.
            If the transaction takes too long, the client connection
            will be closed and the transaction aborted.
        """

        self.storages = storages
        msg = ", ".join(
            ["{}:{}:{}".format(name, storage.isReadOnly() and "RO" or "RW",
                           storage.getName())
             for name, storage in storages.items()])
        log("%s created %s with storages: %s" %
            (self.__class__.__name__, read_only and "RO" or "RW", msg))

        self._lock = Lock()
        self.ssl = ssl  # For dev convenience

        self.read_only = read_only
        self.database = None

        # A list, by server, of at most invalidation_queue_size invalidations.
        # The list is kept in sorted order with the most recent
        # invalidation at the front.  The list never has more than
        # self.invq_bound elements.
        self.invq_bound = invalidation_queue_size
        self.invq = {}

        self.zeo_storages_by_storage_id = {}  # {storage_id -> [ZEOStorage]}
        self.lock_managers = {}  # {storage_id -> LockManager}
        self.stats = {}  # {storage_id -> StorageStats}
        for name, storage in storages.items():
            self._setup_invq(name, storage)
            storage.registerDB(StorageServerDB(self, name))
            if client_conflict_resolution:
                # XXX this may go away later, when storages grow
                # configuration for this.
                storage.tryToResolveConflict = never_resolve_conflict
            self.zeo_storages_by_storage_id[name] = []
            self.stats[name] = stats = StorageStats(
                self.zeo_storages_by_storage_id[name])
            if transaction_timeout is None:
                # An object with no-op methods
                timeout = StubTimeoutThread()
            else:
                timeout = TimeoutThread(transaction_timeout)
                timeout.name = f'TimeoutThread for {name}'
                timeout.start()
            self.lock_managers[name] = LockManager(name, stats, timeout)

        self.invalidation_age = invalidation_age
        self.client_conflict_resolution = client_conflict_resolution

        if addr is not None:
            self.acceptor = Acceptor(self, addr, ssl, msgpack)
            if isinstance(addr, tuple) and addr[0]:
                self.addr = self.acceptor.addr
            else:
                self.addr = addr
            self.loop = self.acceptor.loop
            ZODB.event.notify(Serving(self, address=self.acceptor.addr))

    def create_client_handler(self):
        return ZEOStorage(self, self.read_only)

    def _setup_invq(self, name, storage):
        lastInvalidations = getattr(storage, 'lastInvalidations', None)
        if lastInvalidations is None:
            # Using None below doesn't look right, but the first
            # element in invq is never used.  See get_invalidations.
            # (If it was used, it would generate an error, which would
            # be good. :) Doing this allows clients that were up to
            # date when a server was restarted to pick up transactions
            # it subsequently missed.
            self.invq[name] = [(storage.lastTransaction() or z64, None)]
        else:
            self.invq[name] = list(lastInvalidations(self.invq_bound))
            self.invq[name].reverse()

    def register_connection(self, storage_id, zeo_storage):
        """Internal: register a ZEOStorage with a particular storage.

        This is called by ZEOStorage.register().

        The dictionary self.zeo_storages_by_storage_id maps each
        storage name to a list of current ZEOStorages for that
        storage; this information is needed to handle invalidation.
        This function updates this dictionary.

        Returns the timeout and stats objects for the appropriate storage.
        """
        self.zeo_storages_by_storage_id[storage_id].append(zeo_storage)
        return self.stats[storage_id]

    def _invalidateCache(self, storage_id):
        """We need to invalidate any caches we have.

        This basically means telling our clients to
        invalidate/revalidate their caches. We do this by closing them
        and making them reconnect.
        """

        # This method is called from foreign threads.  We have to
        # worry about interaction with the main thread.

        # Rebuild invq
        self._setup_invq(storage_id, self.storages[storage_id])

        # Make a copy since we are going to be mutating the
        # connections indirectoy by closing them.  We don't care about
        # later transactions since they will have to validate their
        # caches anyway.
        for zs in self.zeo_storages_by_storage_id[storage_id][:]:
            zs.call_soon_threadsafe(zs.connection.close)

    def invalidate(self, zeo_storage, storage_id, tid, invalidated):
        """Internal: broadcast invalidations to clients.

        This is called from several ZEOStorage methods.

        invalidated is a sequence of oids.
        """

        # This method can be called from foreign threads.  We have to
        # worry about interaction with the main thread.

        invq = self.invq[storage_id]
        if len(invq) >= self.invq_bound:
            invq.pop()
        invq.insert(0, (tid, invalidated))

        for zs in self.zeo_storages_by_storage_id[storage_id]:
            if zs is not zeo_storage:
                zs.async_threadsafe('invalidateTransaction', tid, invalidated)

    def broadcast_info(self, storage_id, info):
        """Internal: broadcast info to clients.
        """
        for zs in self.zeo_storages_by_storage_id[storage_id]:
            zs.async_threadsafe('info', info)

    def get_invalidations(self, storage_id, tid):
        """Return a tid and list of all objects invalidation since tid.

        The tid is the most recent transaction id seen by the client.

        Returns None if it is unable to provide a complete list
        of invalidations for tid.  In this case, client should
        do full cache verification.

        XXX This API is stupid.  It would be better to simply return a
        list of oid-tid pairs. With this API, we can't really use the
        tid returned and have to discard all versions for an OID. If
        we used the max tid, then loadBefore results from the cache
        might be incorrect.
        """

        # We make a copy of invq because it might be modified by a
        # foreign (other than main thread) calling invalidate above.
        invq = self.invq[storage_id][:]

        oids = set()
        latest_tid = None
        if invq and invq[-1][0] <= tid:
            # We have needed data in the queue
            for _tid, L in invq:
                if _tid <= tid:
                    break
                oids.update(L)
            latest_tid = invq[0][0]
        elif (self.invalidation_age and
              (self.invalidation_age >
               (time.time()-ZODB.TimeStamp.TimeStamp(tid).timeTime())
               )
              ):
            for t in self.storages[storage_id].iterator(p64(u64(tid)+1)):
                for r in t:
                    oids.add(r.oid)
                latest_tid = t.tid
        elif not invq:
            log("invq empty")
        else:
            log(f"tid to old for invq {u64(tid)} < {u64(invq[-1][0])}")

        return latest_tid, list(oids)

    __thread = None

    def start_thread(self, daemon=True):
        self.__thread = thread = threading.Thread(target=self.loop)
        thread.name = "StorageServer(%s)" % _addr_label(self.addr)
        thread.daemon = daemon
        thread.start()

    __closed = False

    def close(self, join_timeout=1):
        """Close the dispatcher so that there are no new connections.

        This is only called from the test suite, AFAICT.
        """
        if self.__closed:
            return
        self.__closed = True

        # Stop accepting connections
        self.acceptor.close()

        ZODB.event.notify(Closed(self))

        # Close open client connections
        for sid, zeo_storages in self.zeo_storages_by_storage_id.items():
            for zs in zeo_storages[:]:
                try:
                    logger.debug("Closing %s", zs.connection)
                    zs.call_soon_threadsafe(zs.connection.close)
                except Exception:
                    logger.exception("closing connection %r", zs)

        for name, storage in self.storages.items():
            logger.info("closing storage %r", name)
            storage.close()

        if self.__thread is not None:
            self.__thread.join(join_timeout)

        self.acceptor = self.loop = None  # break reference cycles

    def close_conn(self, zeo_storage):
        """Remove the given zeo_storage from self.zeo_storages_by_storage_id.

        This is the inverse of register_connection().
        """
        for zeo_storages in self.zeo_storages_by_storage_id.values():
            if zeo_storage in zeo_storages:
                zeo_storages.remove(zeo_storage)

    def server_status(self, storage_id):
        status = self.stats[storage_id].__dict__.copy()
        status['connections'] = len(status['connections'])
        lock_manager = self.lock_managers[storage_id]
        status['waiting'] = len(lock_manager.waiting)
        status['timeout-thread-is-alive'] = lock_manager.timeout.is_alive()
        last_transaction = self.storages[storage_id].lastTransaction()
        last_transaction_hex = codecs.encode(last_transaction, 'hex_codec')
        # doctests and maybe clients expect a str, not bytes
        last_transaction_hex = str(last_transaction_hex, 'ascii')
        status['last-transaction'] = last_transaction_hex
        return status

    def ruok(self):
        return {storage_id: self.server_status(storage_id)
                    for storage_id in self.storages}


class StubTimeoutThread:

    def begin(self, client):
        pass

    def end(self, client):
        pass

    def is_alive(self):
        return 'stub'


class TimeoutThread(threading.Thread):
    """Monitors transaction progress and generates timeouts."""

    # There is one TimeoutThread per storage, because there's one
    # transaction lock per storage.

    def __init__(self, timeout):
        threading.Thread.__init__(self)
        self.name = "TimeoutThread"
        self.daemon = True
        self._timeout = timeout
        self._client = None
        self._deadline = None
        self._cond = threading.Condition()  # Protects _client and _deadline

    def begin(self, client):
        # Called from the restart code the "main" thread, whenever the
        # storage lock is being acquired.
        with self._cond:
            assert self._client is None
            self._client = client
            self._deadline = time.time() + self._timeout
            self._cond.notify()

    def end(self, client):
        # Called from the "main" thread whenever the storage lock is
        # being released.
        with self._cond:
            assert self._client is not None
            assert self._client is client
            self._client = None
            self._deadline = None

    def run(self):
        # Code running in the thread.
        while 1:
            with self._cond:
                while self._deadline is None:
                    self._cond.wait()
                howlong = self._deadline - time.time()
                if howlong <= 0:
                    # Prevent reporting timeout more than once
                    self._deadline = None
                client = self._client  # For the howlong <= 0 branch below

            if howlong <= 0:
                client.log("Transaction timeout after %s seconds" %
                           self._timeout, logging.CRITICAL)
                try:
                    client.call_soon_threadsafe(client.connection.close)
                except:  # NOQA: E722 bare except
                    client.log("Timeout failure", logging.CRITICAL,
                               exc_info=sys.exc_info())
                    self.end(client)
            else:
                time.sleep(howlong)


def run_in_thread(method, *args):
    t = SlowMethodThread(method, args)
    t.start()
    return t.delay


class SlowMethodThread(threading.Thread):
    """Thread to run potentially slow storage methods.

    Clients can use the delay attribute to access the MTDelay object
    used to send a zrpc response at the right time.
    """

    # Some storage methods can take a long time to complete.  If we
    # run these methods in response to an I/O event, they
    # will block all other server activity until they complete.  To
    # avoid blocking, we spawn a separate thread, return an MTDelay()
    # object, and have the thread reply() when it finishes.

    def __init__(self, method, args):
        threading.Thread.__init__(self)
        self.name = f'SlowMethodThread for {method.__name__}'
        self._method = method
        self._args = args
        self.delay = MTDelay()

    def run(self):
        try:
            result = self._method(*self._args)
        except (SystemExit, KeyboardInterrupt):
            raise
        except Exception:
            self.delay.error(sys.exc_info())
        else:
            self.delay.reply(result)


def _addr_label(addr):
    if isinstance(addr, bytes):
        return addr.decode('ascii')
    if isinstance(addr, str):
        return addr
    else:
        host, port = addr
        return str(host) + ":" + str(port)


class CommitLog:

    def __init__(self):
        self.file = tempfile.TemporaryFile(suffix=".comit-log")
        self.pickler = Pickler(self.file, 1)
        self.pickler.fast = 1
        self.stores = 0

    def size(self):
        return self.file.tell()

    def delete(self, oid, serial):
        self.pickler.dump(('_delete', (oid, serial)))
        self.stores += 1

    def checkread(self, oid, serial):
        self.pickler.dump(('_checkread', (oid, serial)))
        self.stores += 1

    def store(self, oid, serial, data):
        self.pickler.dump(('_store', (oid, serial, data)))
        self.stores += 1

    def restore(self, oid, serial, data, prev_txn):
        self.pickler.dump(('_restore', (oid, serial, data, prev_txn)))
        self.stores += 1

    def undo(self, transaction_id):
        self.pickler.dump(('_undo', (transaction_id, )))
        self.stores += 1

    def __iter__(self):
        self.file.seek(0)
        unpickler = Unpickler(self.file)
        for i in range(self.stores):
            yield unpickler.load()

    def close(self):
        if self.file:
            self.file.close()
            self.file = None


class ServerEvent:

    def __init__(self, server, **kw):
        self.__dict__.update(kw)
        self.server = server


class Serving(ServerEvent):
    pass


class Closed(ServerEvent):
    pass


def never_resolve_conflict(oid, committedSerial, oldSerial, newpickle,
                           committedData=b''):
    raise ConflictError(oid=oid, serials=(committedSerial, oldSerial),
                        data=newpickle)


class LockManager:

    def __init__(self, storage_id, stats, timeout):
        self.storage_id = storage_id
        self.stats = stats
        self.timeout = timeout
        self.locked = None
        self.waiting = {}  # {ZEOStorage -> (func, delay)}
        self._lock = RLock()

    def lock(self, zs, func):
        """Call the given function with the commit lock.

        If we can get the lock right away, return the result of
        calling the function.

        If we can't get the lock right away, return a delay

        The function must set ``locked`` on the zeo-storage to
        indicate that the zeo-storage should be locked.  Otherwise,
        the lock isn't held pas the call.
        """
        with self._lock:
            if self._can_lock(zs):
                self._locked(zs)
            else:
                if any(w for w in self.waiting if w is zs):
                    raise StorageTransactionError("Already voting (waiting)")

                delay = Delay()
                self.waiting[zs] = (func, delay)
                self._log_waiting(
                    zs, "(%r) queue lock: transactions waiting: %s")

                return delay

        try:
            result = func()
        except Exception:
            self.release(zs)
            raise
        else:
            if not zs.locked:
                self.release(zs)
            return result

    def _lock_waiting(self, zs):
        waiting = None
        with self._lock:
            if self.locked is zs:
                assert zs.locked
                return

            if self._can_lock(zs):
                waiting = self.waiting.pop(zs, None)
                if waiting:
                    self._locked(zs)

        if waiting:
            func, delay = waiting
            try:
                result = func()
            except Exception:
                delay.error(sys.exc_info())
                self.release(zs)
            else:
                delay.reply(result)
                if not zs.locked:
                    self.release(zs)

    def release(self, zs):
        with self._lock:
            locked = self.locked
            if locked is zs:
                self._unlocked(zs)

                for zs in list(self.waiting):
                    zs.call_soon_threadsafe(self._lock_waiting, zs)

            else:
                if self.waiting.pop(zs, None):
                    self._log_waiting(
                        zs, "(%r) dequeue lock: transactions waiting: %s")

    def _log_waiting(self, zs, message):
        length = len(self.waiting)
        zs.log(message % (self.storage_id, length),
               logging.CRITICAL if length > 9 else (
                   logging.WARNING if length > 3 else logging.DEBUG)
               )

    def _can_lock(self, zs):
        locked = self.locked

        if locked is zs:
            raise StorageTransactionError("Already voting (locked)")

        if locked is not None:
            if not locked.connected:
                locked.log("Still locked after disconnected. Unlocking.",
                           logging.CRITICAL)
                if locked.transaction:
                    locked.storage.tpc_abort(locked.transaction)

                self._unlocked(locked)
                locked = None

            # Note that locked.locked may not be true here, because
            # .lock may be set in the lock callback, but may not have
            # been set yet.  This aspect of the API may need more
            # thought. :/

        return locked is None

    def _locked(self, zs):
        self.locked = zs
        self.stats.lock_time = time.time()
        self._log_waiting(zs, "(%r) lock: transactions waiting: %s")
        self.timeout.begin(zs)
        return True

    def _unlocked(self, zs):
        assert self.locked is zs
        self.timeout.end(zs)
        self.locked = self.stats.lock_time = None
        zs.locked = False
        self._log_waiting(zs, "(%r) unlock: transactions waiting: %s")



import asyncio
from _thread import allocate_lock
from time import sleep


try:
    ConnectionRefusedError
except NameError:
    class ConnectionRefusedError(OSError):
        pass


class Loop:
    """Simple loop for testing purposes.

    It calls callbacks directly (instead of in the next round);
    it remembers ``call_later`` calls rather than schedule them;
    it does not check calls to non threadsafe methods.
    """

    def __init__(self, addrs=(), debug=True):
        self.protocol = self.transport = None
        self.addrs = addrs
        self.get_debug = lambda: debug
        self.connecting = {}
        self.later = []
        self.exceptions = []

    def call_soon(self, func, *args, **kw):
        # Python 3.7+ calls us with a `context` keyword-only argument:
        kw.pop('context', None)
        assert not kw
        func(*args)

    def _connect(self, future, protocol_factory):
        self.protocol = protocol = protocol_factory()
        self.transport = transport = Transport(protocol)
        protocol.connection_made(transport)
        future.set_result((transport, protocol))

    def connect_connecting(self, addr):
        future, protocol_factory = self.connecting.pop(addr)
        self._connect(future, protocol_factory)

    def fail_connecting(self, addr):
        future, protocol_factory = self.connecting.pop(addr)
        if not future.cancelled():
            future.set_exception(ConnectionRefusedError())

    def create_connection(self, protocol_factory, host=None, port=None,
                          sock=None, ssl=None, server_hostname=None):
        future = asyncio.Future(loop=self)
        if sock is None:
            addr = host, port
            if addr in self.addrs:
                self._connect(future, protocol_factory)
            else:
                self.connecting[addr] = future, protocol_factory
        else:
            self._connect(future, protocol_factory)

        return future

    def create_unix_connection(self, protocol_factory, path):
        future = asyncio.Future(loop=self)
        if path in self.addrs:
            self._connect(future, protocol_factory)
        else:
            self.connecting[path] = future, protocol_factory

        return future

    def call_soon_threadsafe(self, func, *args):
        func(*args)
        return Handle()

    def call_later(self, delay, func, *args):
        handle = Handle()
        self.later.append((delay, func, args, handle))
        return handle

    def call_exception_handler(self, context):
        self.exceptions.append(context)

    closed = False

    def close(self):
        # break reference cycles
        Loop.__init__(self)
        self.closed = True

    def create_future(self):
        return asyncio.Future(loop=self)

    stopped = False

    def stop(self):
        self.stopped = True


AsyncioLoop = asyncio.get_event_loop_policy()._loop_factory


class FaithfulLoop(Loop, AsyncioLoop):
    """Testing loop variant with true ``asyncio`` ``call_*`` methods."""
    def __init__(self, addrs=(), debug=True):
        Loop.__init__(self, addrs, debug)
        AsyncioLoop.__init__(self)

    call_soon = AsyncioLoop.call_soon

    call_soon_threadsafe = AsyncioLoop.call_soon_threadsafe

    def call_later(self, delay, func, *args):
        th = AsyncioLoop.call_later(self, delay, func, *args)
        self.later.append((delay, func, args, th))
        return th

    def call_exception_handler(self, context):
        Loop.call_exception_handler(self, context)
        AsyncioLoop.call_exception_handler(self, context)  # log exception

    def _connect(self, future, protocol_factory):
        """a real ``asyncio`` loop checks that non threadsafe
        functions are only called from the ``asyncio`` context.
        Our protocols assume they are only called from within this
        context. The tests however call their method differently.
        Therefore, we wrap the protocol to properly delegate to the
        ``asyncio`` context.
        """
        protocol = self.protocol = protocol_factory()
        self.transport = transport = Transport(protocol)
        protocol.connection_made(transport)
        protocol = self.protocol = _ProtocolWrapper(self)
        future.set_result((transport, protocol))

    def stop(self):
        AsyncioLoop.stop(self)
        Loop.stop(self)

    def close(self):
        if not self.is_closed():
            AsyncioLoop.close(self)
            Loop.close(self)

    _inactivity_checker_scheduled = False
    _inactivity_lock = allocate_lock()

    def run_until_inactive(self):
        """return when the loop becomes inactive."""
        with self._inactivity_lock:
            if not self._inactivity_checker_scheduled:
                self._inactivity_checker_scheduled = True
                self.call_soon_threadsafe(self._check_inactive)
        while True:
            sleep(0.005)
            with self._inactivity_lock:
                if not self._inactivity_checker_scheduled:
                    return

    def _check_inactive(self):
        """check whether the loop has nothing to do.

        In this case, reset ``_inactivity_checker_scheduled``,
        otherwise, reschedule itself.
        """
        with self._inactivity_lock:
            # We use here implementation details
            if len(self._ready) == 0:  # inactive
                self._inactivity_checker_scheduled = False
            else:
                self.call_soon(self._check_inactive)


class _ProtocolWrapper:
    """Wrapped protocol.

    Protocol instances can only be used by their ``asyncio`` thread.
    The wrapper allows tests (run in a different thread) to
    use the wrapped protocol normally by properly delegating
    method calls to the ``asyncio`` thread.
    """

    def __init__(self, loop):
        self._loop = loop
        self._protocol = loop.protocol

    def data_received(self, data):
        # perform in IO thread
        self._loop.call_soon_threadsafe(self._protocol.data_received, data)
        self._loop.run_until_inactive()

    def connection_lost(self, exc):
        self._loop.call_soon_threadsafe(self._protocol.connection_lost, exc)
        self._loop.run_until_inactive()

    _protocol = None

    def __getattr__(self, attr):
        return getattr(self._protocol, attr)


class Handle:

    _cancelled = False

    def cancel(self):
        self._cancelled = True


class Transport:

    capacity = 1 << 64
    paused = False
    extra = dict(peername='1.2.3.4', sockname=('127.0.0.1', 4200), socket=None)

    def __init__(self, protocol):
        self.data = []
        self.protocol = protocol

    def write(self, data):
        self.data.append(data)
        self.check_pause()

    def writelines(self, lines):
        self.data.extend(lines)
        self.check_pause()

    def check_pause(self):
        if len(self.data) > self.capacity and not self.paused:
            self.paused = True
            self.protocol.pause_writing()

    def pop(self, count=None):
        if count:
            r = self.data[:count]
            del self.data[:count]
        else:
            r = self.data[:]
            del self.data[:]
        self.check_resume()
        return r

    def check_resume(self):
        if len(self.data) < self.capacity and self.paused:
            self.paused = False
            self.protocol.resume_writing()

    closed = False

    def close(self):
        if self.closed:
            return
        self.closed = True
        # honor the ``asyncio.Protocol`` obligation:
        # if ``connection_made`` has been called, there will
        # be exactly one ``connection_lost`` call.
        if not self.protocol.connection_lost_called:
            self.protocol.connection_lost(None)
        self.protocol = None  # break reference cycle

    def get_extra_info(self, name):
        return self.extra[name]


class AsyncRPC:
    """Adapt an asyncio API to an RPC to help hysterical tests
    """
    def __init__(self, api):
        self.api = api

    def __getattr__(self, name):
        return lambda *a, **kw: self.api.call(name, *a, **kw)


class ClientRunner:

    def __init__(self, addr, client, cache, storage, read_only, timeout,
                 **kw):
        self.addr = addr
        self.client = client
        self.cache = cache
        self.storage = storage
        self.read_only = read_only
        self.timeout = timeout,
        for name in kw:
            self.__dict__[name] = kw[name]

    def start(self, wait=True):
        pass

    def call(self, method, *args, **kw):
        return getattr(self, method)(*args)

    async_ = async_iter = call

    def wait(self, timeout=None):
        pass

    def close(self):
        self.cache.close()  # client responsibility



import asyncio
import collections
import logging
import struct
import threading
import unittest
from functools import partial
from itertools import count
from time import sleep
from time import time
from unittest import TestCase
from unittest import mock

from ZODB.utils import RLock
from ZODB.utils import maxtid
from zope.testing import setupstack

from ..Exceptions import ClientDisconnected
from ..Exceptions import ProtocolError
from .base import SizedMessageProtocol
from .base import ZEOBaseProtocol
from .client import ClientThread
from .client import Fallback
from .futures import AsyncTask
from .futures import ConcurrentFuture
from .futures import ConcurrentTask
from .futures import Future
from .marshal import decoder
from .marshal import encoder
from .server import best_protocol_version
from .server import new_connection
from .testing import FaithfulLoop
from .testing import Loop


logger = logging.getLogger(__name__)


class Base:
    enc = b'Z'
    seq_type = list

    def setUp(self):
        super().setUp()
        self.encode = encoder(self.enc)
        self.decode = decoder(self.enc)

    def unsized(self, data, unpickle=False):
        result = []
        while data:
            size, message = data[:2]
            data = data[2:]
            self.assertEqual(struct.unpack(">I", size)[0], len(message))
            if unpickle:
                message = self.decode(message)
            result.append(message)

        if len(result) == 1:
            result = result[0]
        return result

    def parse(self, data):
        return self.unsized(data, True)

    target = None

    def send(self, method, *args, **kw):
        target = kw.pop('target', self.target)
        called = kw.pop('called', True)
        no_output = kw.pop('no_output', True)
        self.assertFalse(kw)

        self.loop.protocol.data_received(
            sized(self.encode(0, True, method, args)))
        if target is not None:
            target = getattr(target, method)
            if called:
                target.assert_called_with(*args)
                target.reset_mock()
            else:
                self.assertFalse(target.called)
        if no_output:
            self.assertFalse(self.loop.transport.pop())

    def pop(self, count=None, parse=True):
        self.assertEqual(self.loop.exceptions, [])
        return self.unsized(self.loop.transport.pop(count), parse)


class ClientTests(Base, setupstack.TestCase, ClientThread):
    """Test ``ClientIO``.

    The tests emulate a server and its responses to verify
    that ``ClientIO`` and ``client.Protocol`` instances behave
    as they should:
    messages sent by those instances can be accessed via ``pop``;
    emulated server actions are generated by calls to ``respond``
    or ``server_async_call``.

    The tests always use an ``asyncio`` loop (even ``uvloop`` tests).
    """

    maxDiff = None

    def tearDown(self):
        self.exit_future_mode()
        self.close()
        super().tearDown()
        loop = self.loop
        if loop is not None:
            self.assertEqual(loop.exceptions, [])
        # ``mock`` creates cyclic structures; break the cycles
        _break_mock_cycles(self.target)

    # For normal operation all (server) interface calls are synchronous:
    # they wait for the result.
    # However, we want to perform our tests without a real server;
    # instead we emulate server responses (via ``respond``)
    # and asynchronous server calls to us via ``send``.
    # For this to work, interface calls must not wait for a result
    # (which we ourselves often must produce).
    # We achieve this by switching to "future mode":
    # in this mode, the interface calls return futures representing
    # the final result. We can call their ``result`` method to
    # wait for their result
    future_mode = None

    def enter_future_mode(self):
        if self.future_mode is None:
            self.future_mode = self.io_call
            # switch for most interface calls to future mode
            # exceptions are those that explicitly use ``wait=True`.
            self.io_call = partial(self.io_call, wait=False)

    def exit_future_mode(self):
        if self.future_mode is not None:
            self.io_call = self.future_mode  # back to synchronous mode
            del self.future_mode

    def start(self,
              addrs=(('127.0.0.1', 8200), ), loop_addrs=None,
              read_only=False,
              finish_start=False,
              future_mode=True,
              ):
        # To create a client, we need to specify an address, a client
        # object and a cache.

        wrapper = mock.Mock(__name__="__name__")
        self.target = wrapper
        cache = MemoryCache()
        # We can also provide an event loop.  We'll use a testing loop
        # so we don't have to actually make any network connection.
        self.loop = loop = FaithfulLoop(
            addrs if loop_addrs is None else loop_addrs)
        ClientThread.__init__(self,
                              addrs, wrapper, cache, 'TEST',
                              read_only, timeout=1)
        loop.run_until_inactive()
        self.assertFalse(wrapper.notify_disconnected.called)
        protocol = loop.protocol
        transport = loop.transport

        if finish_start:
            protocol.data_received(sized(self.enc + b'5'))
            self.assertEqual(self.pop(2, False), self.enc + b'5')
            self.respond(1, None)
            self.respond(2, 'a'*8)
            self.pop(4)
            self.assertEqual(self.pop(), (3, False, 'get_info', ()))
            self.respond(3, dict(length=42))

        if future_mode:
            self.enter_future_mode()
        return (wrapper, cache, self.loop, self.client, protocol, transport)

    def pop(self, *args, **kw):
        self.loop.run_until_inactive()
        return super().pop(*args, **kw)

    def respond(self, message_id, result, async_=False,
                check_exc=True, return_msg=False, protocol=None):
        """emulate a server response."""
        msg = sized(self.encode(message_id, async_, '.reply', result))
        if return_msg:
            return msg
        if protocol is None:
            protocol = self.loop.protocol
        protocol.data_received(msg)
        if check_exc:
            self.assertEqual(self.loop.exceptions, [])

    def server_async_call(self, method, *args,
                          check_exc=True, return_msg=False):
        msg = sized(self.encode(0, True, method, args))
        if return_msg:
            return msg
        self.loop.protocol.data_received(msg)
        if check_exc:
            self.assertEqual(self.loop.exceptions, [])

    def sync_call(self, meth, *args):
        """call future returning *meth* with *args* and wait for result."""
        return meth(*args).result(2)

    def observe(self, f, *args):
        self.observed = f(*args)

    def testClientBasics(self):

        # Here, we'll go through the basic usage of the asyncio ZEO
        # network client.  The client is responsible for the core
        # functionality of a ZEO client storage.  The client storage
        # is largely just a wrapper around the asyncio client.

        wrapper, cache, loop, client, protocol, transport = self.start()
        self.assertFalse(wrapper.notify_disconnected.called)

        # The client isn't connected until the server sends it some data.
        self.assertFalse(client.connected.done() or transport.data)

        # The server sends the client it's protocol. In this case,
        # it's a very high one.  The client will send it's highest that
        # it can use.
        protocol.data_received(sized(self.enc + b'99999'))

        # The client sends back a handshake, and registers the
        # storage, and requests the last transaction.
        self.assertEqual(self.pop(2, False), self.enc + b'5')
        self.assertEqual(loop.exceptions, [])
        self.assertEqual(self.pop(), (1, False, 'register', ('TEST', False)))

        # The client isn't connected until it initializes it's cache:
        self.assertFalse(client.connected.done() or transport.data)

        # If we try to make calls while the client is *initially*
        # connecting, we get an error. This is because some dufus
        # decided to create a client storage without waiting for it to
        # connect.
        self.assertRaises(ClientDisconnected, self.sync_call,
                          self.call, 'foo', 1, 2)

        # When the client is reconnecting, it's ready flag is set to False and
        # it queues calls:
        client.ready = False
        f1 = self.call('foo', 1, 2)
        self.assertFalse(f1.done())

        # If we try to make an async call, we get an immediate error:
        self.assertRaises(ClientDisconnected, self.sync_call,
                          self.async_, 'bar', 3, 4)

        # The wrapper object (ClientStorage) hasn't been notified:
        self.assertFalse(wrapper.notify_connected.called)

        # Let's respond to the register call:
        self.respond(1, None)

        # The client requests the last transaction:
        self.assertEqual(self.pop(), (2, False, 'lastTransaction', ()))

        # We respond
        self.respond(2, b'a'*8)

        # After verification, the client requests info:
        self.assertEqual(self.pop(), (3, False, 'get_info', ()))
        self.respond(3, dict(length=42))

        # Now we're connected, the cache was initialized, and the
        # queued message has been sent:
        self.assertTrue(client.connected.done())
        self.assertEqual(cache.getLastTid(), b'a'*8)
        self.assertEqual(self.pop(), (4, False, 'foo', (1, 2)))

        # The wrapper object (ClientStorage) has been notified:
        wrapper.notify_connected.assert_called_with(client, {'length': 42})

        self.respond(4, 42)
        self.assertEqual(f1.result(), 42)

        # Now we can make async calls:
        self.async_('bar', 3, 4)
        self.assertEqual(self.pop(), (0, True, 'bar', (3, 4)))

        # Loading objects gets special handling to leverage the cache.
        loaded = self.load_before(b'1'*8, maxtid)

        # The data wasn't in the cache, so we made a server call:
        self.assertEqual(self.pop(),
                         ((b'1'*8, maxtid),
                          False,
                          'loadBefore',
                          (b'1'*8, maxtid)))
        # Note load_before uses ``(oid, tid)`` as message id.
        self.respond((b'1'*8, maxtid), (b'data', b'a'*8, None))
        self.assertEqual(loaded.result(), (b'data', b'a'*8, None))

        # If we make another request, it will be satisfied from the cache:
        loaded = self.load_before(b'1'*8, maxtid)
        self.assertEqual(loaded.result(), (b'data', b'a'*8, None))
        self.assertFalse(transport.data)

        # Let's send an invalidation:
        self.send('invalidateTransaction', b'b'*8, self.seq_type([b'1'*8]))

        # Now, if we try to load current again, we'll make a server request.
        loaded = self.load_before(b'1'*8, maxtid)

        # Note that if we make another request for the same object,
        # the requests will be collapsed:
        loaded2 = self.load_before(b'1'*8, maxtid)

        self.assertEqual(self.pop(),
                         ((b'1'*8, maxtid),
                          False,
                          'loadBefore',
                          (b'1'*8, maxtid)))
        self.respond((b'1'*8, maxtid), (b'data2', b'b'*8, None))
        self.assertEqual(loaded.result(), (b'data2', b'b'*8, None))
        self.assertEqual(loaded2.result(), (b'data2', b'b'*8, None))

        # Loading non-current data may also be satisfied from cache
        loaded = self.load_before(b'1'*8, b'b'*8)
        self.assertEqual(loaded.result(), (b'data', b'a'*8, b'b'*8))
        self.assertFalse(transport.data)
        loaded = self.load_before(b'1'*8, b'c'*8)
        self.assertEqual(loaded.result(), (b'data2', b'b'*8, None))
        self.assertFalse(transport.data)
        loaded = self.load_before(b'1'*8, b'_'*8)

        self.assertEqual(self.pop(),
                         ((b'1'*8, b'_'*8),
                          False,
                          'loadBefore',
                          (b'1'*8, b'_'*8)))
        self.respond((b'1'*8, b'_'*8), (b'data0', b'^'*8, b'_'*8))
        self.assertEqual(loaded.result(), (b'data0', b'^'*8, b'_'*8))

        # When committing transactions, we need to update the cache
        # with committed data.  To do this, we pass a (oid, data, resolved)
        # iteratable to tpc_finish_co.

        tids = []

        def finished_cb(tid):
            tids.append(tid)

        committed = self.tpc_finish(
            b'd'*8,
            [(b'2'*8, 'committed 2', False),
             (b'1'*8, 'committed 3', True),
             (b'4'*8, 'committed 4', False),
             ],
            finished_cb)
        self.assertFalse(committed.done() or
                         cache.load(b'2'*8) or
                         cache.load(b'4'*8))
        self.assertEqual(cache.load(b'1'*8), (b'data2', b'b'*8))
        self.assertEqual(self.pop(),
                         (5, False, 'tpc_finish', (b'd'*8,)))
        self.respond(5, b'e'*8)
        self.assertEqual(committed.result(), b'e'*8)
        self.assertEqual(cache.load(b'1'*8), None)
        self.assertEqual(cache.load(b'2'*8), ('committed 2', b'e'*8))
        self.assertEqual(cache.load(b'4'*8), ('committed 4', b'e'*8))
        self.assertEqual(tids.pop(), b'e'*8)

        # If the protocol is disconnected, it will reconnect and will
        # resolve outstanding requests with exceptions:
        loaded = self.load_before(b'1'*8, maxtid)
        f1 = self.call('foo', 1, 2)
        self.assertFalse(loaded.done() or f1.done())
        self.assertEqual(
            self.pop(),
            [((b'11111111', b'\x7f\xff\xff\xff\xff\xff\xff\xff'),
              False, 'loadBefore', (b'1'*8, maxtid)),
             (6, False, 'foo', (1, 2))],
            )
        exc = TypeError(43)

        self.assertFalse(wrapper.notify_disconnected.called)
        wrapper.notify_connected.reset_mock()
        protocol.connection_lost(exc)
        wrapper.notify_disconnected.assert_called_with()

        self.assertIsInstance(loaded.exception(), ClientDisconnected)
        self.assertEqual(loaded.exception().args, (exc,))
        self.assertIsInstance(f1.exception(), ClientDisconnected)
        self.assertEqual(f1.exception().args, (exc,))

        # Because we reconnected, a new protocol and transport were created:
        self.assertIsNot(protocol, loop.protocol)
        self.assertTrue(protocol.closed)
        self.assertIsNot(transport, loop.transport)
        self.assertTrue(transport.closed)
        protocol = loop.protocol
        transport = loop.transport

        # and we have a new incomplete connect future:
        self.assertFalse(client.connected.done() or transport.data)

        # This time we'll send a lower protocol version.  The client
        # will send it back, because it's lower than the client's
        # protocol:
        protocol.data_received(sized(self.enc + b'5'))
        self.assertEqual(self.unsized(transport.pop(2)), self.enc + b'5')
        self.assertEqual(self.pop(), (1, False, 'register', ('TEST', False)))
        self.assertFalse(wrapper.notify_connected.called)

        # If the register response is a tid, then the client won't
        # request lastTransaction
        self.respond(1, b'e'*8)
        self.assertEqual(self.pop(), (2, False, 'get_info', ()))
        self.respond(2, dict(length=42))

        # Because the server tid matches the cache tid, we're done connecting
        wrapper.notify_connected.assert_called_with(client, {'length': 42})
        self.assertTrue(client.connected.done() and not transport.data)
        self.assertEqual(cache.getLastTid(), b'e'*8)

        # Because we were able to update the cache, we didn't have to
        # invalidate the database cache:
        self.assertFalse(wrapper.invalidateTransaction.called)

        # The close method closes the connection and cache:
        loop.call_soon_threadsafe(client.close)
        loop.run_until_inactive()
        self.assertTrue(transport.closed and cache.closed and protocol.closed)

        # The client doesn't reconnect
        self.assertEqual(loop.protocol, protocol)
        self.assertEqual(loop.transport, transport)

    def test_cache_behind(self):
        wrapper, cache, loop, client, protocol, transport = self.start()

        cache.setLastTid(b'a'*8)
        cache.store(b'4'*8, b'a'*8, None, '4 data')
        cache.store(b'2'*8, b'a'*8, None, '2 data')

        self.assertFalse(client.connected.done() or transport.data)
        protocol.data_received(sized(self.enc + b'5'))
        self.assertEqual(self.unsized(transport.pop(2)), self.enc + b'5')
        self.respond(1, None)
        self.respond(2, b'e'*8)
        self.pop(4)

        # We have to verify the cache, so we're not done connecting:
        self.assertFalse(client.connected.done())
        self.assertEqual(self.pop(),
                         (3, False, 'getInvalidations', (b'a'*8, )))
        self.respond(3, (b'e'*8, [b'4'*8]))

        self.assertEqual(self.pop(), (4, False, 'get_info', ()))
        self.respond(4, dict(length=42))

        # Now that verification is done, we're done connecting
        self.assertTrue(client.connected.done() and not transport.data)
        self.assertEqual(cache.getLastTid(), b'e'*8)

        # And the cache has been updated:
        self.assertEqual(cache.load(b'2'*8),
                         ('2 data', b'a'*8))  # unchanged
        self.assertEqual(cache.load(b'4'*8), None)

        # Because we were able to update the cache, we didn't have to
        # invalidate the database cache:
        self.assertFalse(wrapper.invalidateCache.called)

    def test_cache_way_behind(self):
        wrapper, cache, loop, client, protocol, transport = self.start()

        cache.setLastTid(b'a'*8)
        cache.store(b'4'*8, b'a'*8, None, '4 data')
        self.assertTrue(cache)

        self.assertFalse(client.connected.done() or transport.data)
        protocol.data_received(sized(self.enc + b'5'))
        self.assertEqual(self.unsized(transport.pop(2)), self.enc + b'5')
        self.respond(1, None)
        self.respond(2, b'e'*8)
        self.pop(4)

        # We have to verify the cache, so we're not done connecting:
        self.assertFalse(client.connected.done())
        self.assertEqual(self.pop(),
                         (3, False, 'getInvalidations', (b'a'*8, )))

        # We respond None, indicating that we're too far out of date:
        self.respond(3, None)

        self.assertEqual(self.pop(), (4, False, 'get_info', ()))
        self.respond(4, dict(length=42))

        # Now that verification is done, we're done connecting
        self.assertTrue(client.connected.done() and not transport.data)
        self.assertEqual(cache.getLastTid(), b'e'*8)

        # But the cache is now empty and we invalidated the database cache
        self.assertFalse(cache)
        wrapper.invalidateCache.assert_called_with()

    def test_multiple_addresses(self):
        # We can pass multiple addresses to client constructor
        addrs = [('1.2.3.4', 8200), ('2.2.3.4', 8200)]
        wrapper, cache, loop, client, protocol, transport = self.start(
            addrs, ())

        # We haven't connected yet
        self.assertIsNone(protocol)
        self.assertIsNone(transport)

        # There are 2 connection attempts outstanding:
        self.loop.run_until_inactive()
        self.assertEqual(sorted(loop.connecting), addrs)

        # We cause the first one to fail:
        loop.call_soon_threadsafe(loop.fail_connecting, addrs[0])
        self.loop.run_until_inactive()
        self.assertEqual(sorted(loop.connecting), addrs[1:])

        # The failed connection is attempted in the future:
        delay, func, args, handle = loop.later.pop(0)
        self.assertTrue(1 <= delay <= 2)
        handle.cancel()
        loop.call_soon_threadsafe(func, *args)
        self.loop.run_until_inactive()
        self.assertEqual(sorted(loop.connecting), addrs)

        # Let's connect the second address
        loop.call_soon_threadsafe(loop.connect_connecting, addrs[1])
        self.loop.run_until_inactive()
        self.assertEqual(sorted(loop.connecting), addrs[:1])
        protocol = loop.protocol
        transport = loop.transport
        protocol.data_received(sized(self.enc + b'5'))
        self.assertEqual(self.unsized(transport.pop(2)), self.enc + b'5')
        # cancel the heartbeat to make debugging easier
        protocol.heartbeat_handle.cancel()
        self.respond(1, None)

        # Now, when the first connection fails, it won't be retried,
        # because we're already connected.
        # (first in later is heartbeat)
        self.assertEqual(sorted(loop.later[1:]), [])
        loop.call_soon_threadsafe(loop.fail_connecting, addrs[0])
        self.loop.run_until_inactive()
        self.assertEqual(sorted(loop.connecting), [])
        self.assertEqual(sorted(loop.later[1:]), [])

    def test_bad_server_tid(self):
        # If in verification we get a server_tid behind the cache's, make sure
        # we retry the connection later.
        wrapper, cache, loop, client, protocol, transport = self.start()
        cache.store(b'4'*8, b'a'*8, None, '4 data')
        cache.setLastTid('b'*8)
        protocol.data_received(sized(self.enc + b'5'))
        self.assertEqual(self.unsized(transport.pop(2)), self.enc + b'5')
        self.respond(1, None)
        self.respond(2, 'a'*8)
        self.pop()
        self.assertFalse(client.connected.done() or transport.data)
        delay, func, args, _ = loop.later.pop(1)  # first in later is heartbeat
        self.assertTrue(8 < delay < 10)
        self.assertEqual(len(loop.later), 1)  # first in later is heartbeat
        self.loop.call_soon_threadsafe(func, *args)  # connect again
        self.loop.run_until_inactive()
        self.assertIsNot(protocol, loop.protocol)
        self.assertIsNot(transport, loop.transport)
        protocol = loop.protocol
        transport = loop.transport
        protocol.data_received(sized(self.enc + b'5'))
        self.assertEqual(self.unsized(transport.pop(2)), self.enc + b'5')
        self.respond(1, None)
        self.respond(2, 'b'*8)
        self.pop(4)
        self.assertEqual(self.pop(), (3, False, 'get_info', ()))
        self.respond(3, dict(length=42))
        self.assertTrue(client.connected.done() and not transport.data)
        self.assertTrue(client.ready)

    def test_readonly_fallback(self):
        addrs = [('1.2.3.4', 8200), ('2.2.3.4', 8200)]
        wrapper, cache, loop, client, protocol, transport = self.start(
            addrs, (), read_only=Fallback)

        self.assertTrue(self.is_read_only())

        # We'll treat the first address as read-only and we'll let it connect:
        loop.call_soon_threadsafe(loop.connect_connecting, addrs[0])
        self.loop.run_until_inactive()
        protocol, transport = loop.protocol, loop.transport
        protocol.data_received(sized(self.enc + b'5'))
        self.assertEqual(self.unsized(transport.pop(2)), self.enc + b'5')
        # We see that the client tried a writable connection:
        self.assertEqual(self.pop(),
                         (1, False, 'register', ('TEST', False)))
        # We respond with a read-only exception:
        self.respond(1, ('ZODB.POSException.ReadOnlyError', ()), True)
        self.assertTrue(self.is_read_only())

        # The client tries for a read-only connection:
        self.assertEqual(self.pop(), (2, False, 'register', ('TEST', True)))
        # We respond with successfully:
        self.respond(2, None)
        self.pop(2)
        self.respond(3, 'b'*8)
        self.assertTrue(self.is_read_only())

        # At this point, the client is ready and using the protocol,
        # and the protocol is read-only:
        self.assertTrue(client.ready)
        self.assertEqual(client.protocol, protocol._protocol)
        self.assertEqual(protocol.read_only, True)
        connected = client.connected

        # The client asks for info, and we respond:
        self.assertEqual(self.pop(), (4, False, 'get_info', ()))
        self.respond(4, dict(length=42))

        self.assertTrue(connected.done())

        # We connect the second address:
        loop.call_soon_threadsafe(loop.connect_connecting, addrs[1])
        self.loop.run_until_inactive()
        loop.protocol.data_received(sized(self.enc + b'5'))
        self.assertEqual(self.unsized(loop.transport.pop(2)),
                         self.enc + b'5')
        self.assertEqual(self.parse(loop.transport.pop()),
                         (1, False, 'register', ('TEST', False)))
        self.assertTrue(self.is_read_only())

        # We respond and the writable connection succeeds:
        self.respond(1, None)

        # at this point, a lastTransaction request is emitted:

        self.assertEqual(self.parse(loop.transport.pop()),
                         (2, False, 'lastTransaction', ()))
        self.assertFalse(self.is_read_only())

        # Now, the original protocol is closed, and the client is
        # no-longer ready:
        self.assertFalse(client.ready)
        self.assertIsNot(client.protocol, protocol._protocol)
        self.assertEqual(client.protocol, loop.protocol._protocol)
        self.assertEqual(protocol.closed, True)
        self.assertIsNot(client.connected, connected)
        self.assertFalse(client.connected.done())
        protocol, transport = loop.protocol, loop.transport
        self.assertEqual(protocol.read_only, False)

        # Now, we finish verification
        self.respond(2, 'b'*8)
        self.respond(3, dict(length=42))
        self.assertTrue(client.ready)
        self.assertTrue(client.connected.done())

    def test_invalidations_while_verifying(self):
        # While we're verifying, invalidations are ignored
        wrapper, cache, loop, client, protocol, transport = self.start()
        protocol.data_received(sized(self.enc + b'5'))
        self.assertEqual(self.unsized(transport.pop(2)), self.enc + b'5')
        self.respond(1, None)
        self.pop(4)
        self.send('invalidateTransaction', b'b'*8, [b'1'*8], called=False)
        self.respond(2, b'a'*8)
        self.send('invalidateTransaction', b'c'*8, self.seq_type([b'1'*8]),
                  no_output=False)
        self.assertEqual(self.pop(), (3, False, 'get_info', ()))

        # We'll disconnect:
        protocol.connection_lost(Exception("lost"))
        self.assertIsNot(protocol, loop.protocol)
        self.assertIsNot(transport, loop.transport)
        protocol = loop.protocol
        transport = loop.transport

        # Similarly, invalidations aren't processed while reconnecting:

        protocol.data_received(sized(self.enc + b'5'))
        self.assertEqual(self.unsized(transport.pop(2)), self.enc + b'5')
        self.respond(1, None)
        self.pop(4)
        self.send('invalidateTransaction', b'd'*8, [b'1'*8], called=False)
        self.respond(2, b'c'*8)
        self.send('invalidateTransaction', b'e'*8, self.seq_type([b'1'*8]),
                  no_output=False)
        self.assertEqual(self.pop(), (3, False, 'get_info', ()))

    def test_flow_control(self):
        # When sending a lot of data (blobs), we don't want to fill up
        # memory behind a slow socket. Asycio's flow control helper
        # seems a bit complicated. We'd rather pass an iterator that's
        # consumed as we can.

        wrapper, cache, loop, client, protocol, transport = self.start(
            finish_start=True)

        # Give the transport a small capacity:
        transport.capacity = 2
        self.async_('foo')
        self.async_('bar')
        self.async_('baz')
        self.async_('splat')

        # The first 2 were sent, but the remaining were queued.
        self.assertEqual(self.pop(),
                         [(0, True, 'foo', ()), (0, True, 'bar', ())])

        # But popping them allowed sending to resume:
        self.assertEqual(self.pop(),
                         [(0, True, 'baz', ()), (0, True, 'splat', ())])

        # This is especially handy with iterators:
        self.async_iter((name, ()) for name in 'abcde')
        self.async_('f')
        self.assertEqual(self.pop(), [(0, True, 'a', ()), (0, True, 'b', ())])
        self.assertEqual(self.pop(), [(0, True, 'c', ()), (0, True, 'd', ())])
        self.assertEqual(self.pop(), [(0, True, 'e', ()), (0, True, 'f', ())])
        self.assertEqual(self.pop(), [])

    def test_bad_protocol(self):
        wrapper, cache, loop, client, protocol, transport = self.start()
        with mock.patch("ZEO.asyncio.client.logger.error") as error:
            self.assertFalse(error.called)
            protocol.data_received(sized(self.enc + b'200'))
            self.assertIsInstance(error.call_args[0][1], ProtocolError)

    def test_get_peername(self):
        wrapper, cache, loop, client, protocol, transport = self.start(
            finish_start=True)
        self.assertEqual(client.get_peername(), '1.2.3.4')

    def test_call_async(self):
        # There are a few (1?) cases where we call into client storage
        # where it needs to call back asyncronously. Because we're
        # calling from the same thread, we don't need to use a futurte.
        wrapper, cache, loop, client, protocol, transport = self.start(
            finish_start=True)

        client.call_async('foo', (1, ))
        self.assertEqual(self.pop(), (0, True, 'foo', (1, )))

    def test_ClientDisconnected_on_call_timeout(self):
        (wrapper, cache, loop, client, protocol,
         transport) = self.start(future_mode=False)
        self.assertRaises(ClientDisconnected, self.call, 'foo')
        client.ready = False
        self.assertRaises(ClientDisconnected, self.call, 'foo')

    def test_errors_in_data_received(self):
        # There was a bug in ZEO.async.client.Protocol.data_recieved
        # that caused it to fail badly if errors were raised while
        # handling data.

        wrapper, cache, loop, client, protocol, transport = self.start(
            finish_start=True)

        wrapper.receiveBlobStart.side_effect = ValueError('test')

        chunk = 'x' * 99999
        try:
            loop.protocol.data_received(
                sized(
                    self.encode(0, True, 'receiveBlobStart', ('oid', 'serial'))
                    ) +
                sized(
                    self.encode(
                        0, True, 'receiveBlobChunk', ('oid', 'serial', chunk))
                    )
                )
        except ValueError:
            pass
        loop.protocol.data_received(sized(
            self.encode(0, True, 'receiveBlobStop', ('oid', 'serial'))
            ))
        wrapper.receiveBlobChunk.assert_called_with('oid', 'serial', chunk)
        wrapper.receiveBlobStop.assert_called_with('oid', 'serial')

    def test_heartbeat(self):
        # Protocols run heartbeats on a configurable (sort of)
        # heartbeat interval, which defaults to every 60 seconds.
        wrapper, cache, loop, client, protocol, transport = self.start(
            finish_start=True)

        delay, func, args, handle = loop.later.pop()
        self.assertEqual(
            (delay, func, args, handle),
            (60, protocol.heartbeat, (), protocol.heartbeat_handle),
            )
        self.assertFalse(loop.later or handle._cancelled)

        # The heartbeat function sends heartbeat data and reschedules itself.
        self.loop.call_soon_threadsafe(func)
        self.loop.run_until_inactive()
        self.assertEqual(self.pop(), (-1, 0, '.reply', None))
        self.assertTrue(protocol.heartbeat_handle != handle)

        delay, func, args, handle = loop.later.pop()
        self.assertEqual(
            (delay, func, args, handle),
            (60, protocol.heartbeat, (), protocol.heartbeat_handle),
            )
        self.assertFalse(loop.later or handle._cancelled)

        # The heartbeat is cancelled when the protocol connection is lost:
        protocol.connection_lost(None)
        self.assertTrue(handle._cancelled)

    def test_asyncall_doesnt_overtake_reply(self):
        """verify that an asynchronous call after a reply does not
        become effective before the reply."""
        storage_mock, cache, loop, io, protocol, transport = self.start(
            finish_start=True)
        # our syncchronous call
        f = self.tpc_finish(None, (), lambda *args: None, wait=False)
        self.assertEqual(self.pop(), (4, False, "tpc_finish", (None,)))
        # emulate the reply, followed by an asynchronous call
        # of ``invalidateTransaction``.
        # It is important that the two messages are delivered together
        msg = (self.respond(4, "b"*8, return_msg=True) +
               self.server_async_call("invalidateTransaction", "c"*8, (),
                                      return_msg=True))
        protocol.data_received(msg)
        self.assertEqual(loop.exceptions, [])
        f.result()  # no exception
        self.assertEqual(cache.getLastTid(), "c"*8)

    def test_reply_doesnt_overtake_asyncall(self):
        """verify that a reply does not overtake an asynchronous call."""
        storage_mock, cache, loop, io, protocol, transport = self.start(
            finish_start=True)
        # our synchronous call
        f = self.call("sync", wait=False)
        self.assertEqual(self.pop(), (4, False, "sync", ()))
        # emulate an asynchronous call followed by the reply
        self.observed = None

        def observer():
            """used as the storage's `receiveBlobStop`.
            It records the state of ``f`` and therefore allows
            to check what is processed first.
            """
            self.observed = bool(f.done())

        io.receiveBlobStop = observer
        # Emulate an async call to the client followed by the reply.
        # It is important that the two messages are delivered together.
        msg = (self.server_async_call("receiveBlobStop", return_msg=True) +
               self.respond(4, None, return_msg=True))
        protocol.data_received(msg)
        self.assertEqual(loop.exceptions, [])
        self.assertIs(self.observed, False)

    def test_serialized_registration(self):
        addrs = [('1.2.3.4', 8200), ('2.2.3.4', 8200)]
        wrapper, cache, loop, io, protocol, transport = self.start(
            addrs, ())
        # connect to the first address
        loop.call_soon_threadsafe(loop.connect_connecting, addrs[0])
        self.loop.run_until_inactive()
        protocol_1 = loop.protocol
        transport_1 = loop.transport
        # connect to the second address
        loop.call_soon_threadsafe(loop.connect_connecting, addrs[1])
        self.loop.run_until_inactive()
        protocol_2 = loop.protocol
        transport_2 = loop.transport
        # verify not locked
        self.assertFalse(io.register_lock.locked())
        # start the first protocol
        protocol_1.data_received(sized(self.enc + b'5'))
        self.assertTrue(io.register_lock.locked())
        self.assertEqual(self.unsized(transport_1.pop(2)), self.enc + b'5')
        self.assertEqual(len(transport_1.data), 2)  # register call
        # start the second protocol
        protocol_2.data_received(sized(self.enc + b'5'))
        self.assertEqual(self.unsized(transport_2.pop(2)), self.enc + b'5')
        self.assertEqual(len(transport_2.data), 0)  # waits for lock
        self.assertFalse(io.connected.done())
        # finish first protocol connection
        self.respond(1, None, protocol=protocol_1)  # register
        self.respond(2, b"a"*8, protocol=protocol_1)  # lastTransaction
        self.assertTrue(io.ready)
        self.assertFalse(io.connected.done())
        self.respond(3, {}, protocol=protocol_1)  # get_info
        # the second protocol has got the lock but released it immediately
        # because it was closed due to the accepted protocol 1
        self.assertFalse(io.register_lock.locked())
        self.assertTrue(protocol_2._protocol.closed)
        self.assertTrue(io.ready)
        self.assertTrue(io.connected.done())

    def test_close_with_running_loop(self):
        storage_mock, cache, loop, io, protocol, transport = self.start(
            finish_start=True, future_mode=False)
        self.close()
        self.assertFalse(io.ready)
        self.assertTrue(self.is_closed())
        self.assertTrue(loop.is_closed())
        self.close()  # subsequent calls do not have an effect

    def test_close_with_stopped_loop(self):
        storage_mock, cache, loop, io, protocol, transport = self.start(
            finish_start=True, future_mode=False)
        loop.call_soon_threadsafe(loop.stop)  # schedule loop stop
        # risks race condition as the loop is going to be closed during
        # stop processing
        # loop.run_until_inactive()
        sleep(0.1)  # give the loop time to stop
        self.close()
        self.assertFalse(io.ready)
        self.assertTrue(self.is_closed())
        self.assertTrue(loop.is_closed())
        self.close()  # subsequent calls do not have an effect

    def test_io_close(self):
        storage_mock, cache, loop, io, protocol, transport = self.start(
            finish_start=True)
        tf = self.call("test")
        loop.call_soon_threadsafe(self.observe, io.close)
        loop.run_until_inactive()
        self.assertTrue(self.observed.done())
        self.assertTrue(protocol.closed)
        self.assertTrue(tf.done())
        self.assertIsInstance(tf.exception(), ClientDisconnected)

    def test_io_close_after_connection_loss(self):
        storage_mock, cache, loop, io, protocol, transport = self.start(
            finish_start=True)
        tf1 = self.call("test")
        loop.run_until_inactive()
        # With the following 2 lines, ``connection_lost`` will
        # effectively overtake the call: i.e. it appears as
        # if the connection were lost before the call.
        # This means that the call will start waiting for a reconnection.
        # Further down, we verify that ``close`` cleans up the state;
        # this is necessary as the closing prevents a reconnection.
        tf2 = self.call("test")
        protocol.connection_lost(None)
        loop.call_soon_threadsafe(self.observe, io.close)
        loop.run_until_inactive()
        self.assertTrue(self.observed.done())
        self.assertTrue(protocol.closed)
        self.assertTrue(tf1.done())
        self.assertIsInstance(tf1.exception(), ClientDisconnected)
        self.assertTrue(tf2.done())
        self.assertIsInstance(tf2.exception(), ClientDisconnected)

    def test_io_close_before_connection(self):
        storage_mock, cache, loop, io, protocol, transport = self.start(
            loop_addrs=(),  # prevent auto connection
            finish_start=False)
        loop.call_soon_threadsafe(self.observe, io.close)
        loop.run_until_inactive()
        self.assertTrue(self.observed.done())

    def test_io_close_after_register_exception(self):
        storage_mock, cache, loop, io, protocol, transport = self.start(
            finish_start=False)
        protocol.data_received(sized(self.enc + b'5'))
        # let the register call fail
        self.respond(1, ("Exception", "register_failed"), async_=True)
        loop.call_soon_threadsafe(self.observe, io.close)
        loop.run_until_inactive()
        self.assertTrue(self.observed.done())
        self.assertTrue(protocol.closed)

    def test_io_close_after_register_exception_before_reconnection(self):
        addr = ('127.0.0.1', 8200)
        storage_mock, cache, loop, io, protocol, transport = self.start(
            addrs=(addr,),
            loop_addrs=(),  # prevent auto connection
            finish_start=False)
        loop.call_soon_threadsafe(loop.connect_connecting, addr)  # connect
        loop.run_until_inactive()
        loop.protocol.data_received(sized(self.enc + b'5'))
        # let the register call fail
        self.respond(1, ("Exception", "register_failed"), async_=True)
        # check "reconnection called for"
        try_reconnecting = loop.later[1]  # entry 0 is "heartbeat"
        connect = try_reconnecting[1]
        self.assertEqual(connect.__func__.__name__,
                         "try_connecting")
        loop.call_soon_threadsafe(connect)
        loop.call_soon_threadsafe(self.observe, io.close)
        loop.run_until_inactive()
        self.assertTrue(self.observed.done())
        self.assertTrue(loop.protocol.closed)

    def test_io_close_after_register_exception_after_reconnection(self):
        addr = ('127.0.0.1', 8200)
        storage_mock, cache, loop, io, protocol, transport = self.start(
            addrs=(addr,),
            loop_addrs=(),  # prevent auto connection
            finish_start=False)
        loop.call_soon_threadsafe(loop.connect_connecting, addr)  # connect
        loop.run_until_inactive()
        loop.protocol.data_received(sized(self.enc + b'5'))
        # let the register call fail
        self.respond(1, ("Exception", "register_failed"), async_=True)
        # check "reconnection called for"
        try_reconnecting = loop.later[1]  # entry 0 is "heartbeat"
        connect = try_reconnecting[1]
        self.assertEqual(connect.__func__.__name__,
                         "try_connecting")
        loop.call_soon_threadsafe(connect)
        loop.call_soon_threadsafe(loop.connect_connecting, addr)  # connect
        loop.run_until_inactive()
        loop.protocol.data_received(sized(self.enc + b'5'))
        loop.call_soon_threadsafe(self.observe, io.close)
        loop.run_until_inactive()
        self.assertTrue(self.observed.done())
        self.assertTrue(loop.protocol.closed)

    def test_io_close_after_register_exception_after_connection_lost(self):
        addr = ('127.0.0.1', 8200)
        storage_mock, cache, loop, io, protocol, transport = self.start(
            addrs=(addr,),
            loop_addrs=(),  # prevent auto connection
            finish_start=False)
        loop.call_soon_threadsafe(loop.connect_connecting, addr)  # connect
        loop.run_until_inactive()
        loop.protocol.data_received(sized(self.enc + b'5'))
        # let the register call fail
        self.respond(1, ("Exception", "register_failed"), async_=True)
        loop.protocol.connection_lost("disconnected")
        loop.call_soon_threadsafe(self.observe, io.close)
        loop.run_until_inactive()
        self.assertTrue(self.observed.done())
        self.assertTrue(loop.protocol.closed)


class MsgpackClientTests(ClientTests):
    enc = b'M'
    seq_type = tuple


class MemoryCache:

    def __init__(self):
        # { oid -> [(start, end, data)] }
        self.data = collections.defaultdict(list)
        self.last_tid = None
        self._lock = RLock()

    clear = __init__

    closed = False

    def close(self):
        self.closed = True

    def __len__(self):
        return len(self.data)

    def load(self, oid):
        revisions = self.data[oid]
        if revisions:
            start, end, data = revisions[-1]
            if not end:
                return data, start
        return None

    def store(self, oid, start_tid, end_tid, data):
        assert start_tid is not None
        revisions = self.data[oid]
        data = (start_tid, end_tid, data)
        if not revisions or data != revisions[-1]:
            revisions.append(data)
        revisions.sort()

    def loadBefore(self, oid, tid):
        for start, end, data in self.data[oid]:
            if start < tid and (end is None or end >= tid):
                return data, start, end

    def invalidate(self, oid, tid):
        revisions = self.data[oid]
        if revisions:
            if tid is None:
                del revisions[:]
            else:
                start, end, data = revisions[-1]
                if end is None:
                    revisions[-1] = start, tid, data

    def getLastTid(self):
        return self.last_tid

    def setLastTid(self, tid):
        if self.last_tid is not None and tid < self.last_tid:
            raise ValueError("tids must increase")
        self.last_tid = tid


class ServerTests(Base, setupstack.TestCase):

    # The server side of things is pretty simple compared to the
    # client, because it's the clien't job to make and keep
    # connections. Servers are pretty passive.

    def connect(self, finish=False):
        protocol = server_protocol(self.enc == b'M')
        self.loop = protocol.loop
        self.target = protocol.zeo_storage
        if finish:
            self.assertEqual(self.pop(parse=False),
                             self.enc + best_protocol_version)
            protocol.data_received(sized(self.enc + b'5'))
        return protocol

    message_id = 0
    target = None

    def call(self, meth, *args, **kw):
        if kw:
            expect = kw.pop('expect', self)
            target = kw.pop('target', self.target)
            self.assertFalse(kw)

            if target is not None:
                target = getattr(target, meth)
                if expect is not self:
                    target.return_value = expect

        self.message_id += 1
        self.loop.protocol.data_received(
            sized(self.encode(self.message_id, False, meth, args)))

        if target is not None:
            target.assert_called_once_with(*args)
            target.reset_mock()

        if expect is not self:
            self.assertEqual(self.pop(),
                             (self.message_id, False, '.reply', expect))

    def testServerBasics(self):
        # A simple listening thread accepts connections.  It creats
        # asyncio connections by calling ZEO.asyncio.new_connection:
        protocol = self.connect()
        self.assertFalse(protocol.zeo_storage.notify_connected.called)

        # The server sends it's protocol.
        self.assertEqual(self.pop(parse=False),
                         self.enc + best_protocol_version)

        # The client sends it's protocol:
        protocol.data_received(sized(self.enc + b'5'))

        self.assertEqual(protocol.protocol_version, self.enc + b'5')

        protocol.zeo_storage.notify_connected.assert_called_once_with(protocol)

        # The client registers:
        self.call('register', False, expect=None)

        # It does other things, like, send hearbeats:
        protocol.data_received(sized(self.encode(-1, 0, '.reply', None)))

        # The client can make async calls:
        self.send('register')

        # Let's close the connection
        self.assertFalse(protocol.zeo_storage.notify_disconnected.called)
        protocol.connection_lost(None)
        protocol.zeo_storage.notify_disconnected.assert_called_once_with()

    def test_invalid_methods(self):
        protocol = self.connect(True)
        protocol.zeo_storage.notify_connected.assert_called_once_with(protocol)

        # If we try to call a method that isn't in the protocol's
        # white list, it will disconnect:
        self.assertFalse(protocol.loop.transport.closed)
        self.call('foo', target=None)
        self.assertTrue(protocol.loop.transport.closed)


class MsgpackServerTests(ServerTests):
    enc = b'M'
    seq_type = tuple


def server_protocol(msgpack,
                    zeo_storage=None,
                    protocol_version=None,
                    addr=('1.2.3.4', '42'),
                    ):
    if zeo_storage is None:
        zeo_storage = mock.Mock()
    loop = Loop()
    sock = ()  # anything not None
    new_connection(loop, addr, sock, zeo_storage, msgpack)
    if protocol_version:
        loop.protocol.data_received(sized(protocol_version))
    return loop.protocol


def sized(message):
    return struct.pack(">I", len(message)) + message


class Logging:

    def __init__(self, level=logging.ERROR):
        self.level = level

    def __enter__(self):
        self.handler = logging.StreamHandler()
        logging.getLogger().addHandler(self.handler)
        logging.getLogger().setLevel(self.level)

    def __exit__(self, *args):
        logging.getLogger().removeHandler(self.handler)
        logging.getLogger().setLevel(logging.NOTSET)


class ZEOBaseProtocolTests(setupstack.TestCase):

    def setUp(self):
        self.loop = loop = Loop()
        loop.create_connection(lambda: ZEOBaseProtocol(loop, "proto"),
                               sock=True)

    def tearDown(self):
        self.loop.protocol.close()
        self.loop.close()

    def test_write_message_iter(self):
        """test https://github.com/zopefoundation/ZEO/issues/150."""
        loop = self.loop
        protocol, transport = loop.protocol, loop.transport
        transport.capacity = 1  # single message

        def it(tag):
            yield tag
            yield tag

        protocol.write_message_iter(it(b"0"))
        protocol.write_message_iter(it(b"1"))

        for b in b"0011":
            l, t = transport.pop(2)
            self.assertEqual(l, b"\x00\x00\x00\x01")
            self.assertEqual(t, to_byte(b))


    def test_repr(self):
        repr(self.loop.protocol)  # satisfied if no exception


class SizedMessageProtocolTests(setupstack.TestCase):
    def setUp(self):
        self.loop = loop = Loop()
        self.received = received = []
        loop.create_connection(
            lambda: SizedMessageProtocol(received.append), sock=True)
        self.transport = loop.transport
        self.protocol = loop.protocol

    def tearDown(self):
        self.protocol.close()
        self.loop.close()

    def test_sm_write_message(self):
        protocol, transport = self.protocol, self.transport
        for i in range(2):
            protocol.write_message(to_byte(i))
        for i in range(2):
            l, b = transport.pop(2)
            self.assertEqual(l, b"\x00\x00\x00\x01")
            self.assertEqual(b, to_byte(i))
        protocol.write_message(b"12")
        l, m = transport.pop(2)
        self.assertEqual(l, b"\x00\x00\x00\x02")
        self.assertEqual(m, b"12")

    def test_sm_write_iter(self):
        protocol, transport = self.protocol, self.transport
        protocol.write_message_iter(to_byte(b) for b in b"0123")
        for b in b"0123":
            l, t = transport.pop(2)
            self.assertEqual(l, b"\x00\x00\x00\x01")
            self.assertEqual(t, to_byte(b))

    def test_sm_receive(self):
        protocol = self.protocol
        msgs = b"0 11 222".split()
        for msg in msgs:
            protocol.data_received(sized(msg))
        self.assertEqual(self.received, msgs)

    def test_sm_data_received(self):
        msg = b"a test message"
        data = sized(msg) * 2
        receive = self.protocol.data_received
        expected = [msg] * 2

        def check(*split_size):
            idx = 0
            for size in split_size:
                nidx = idx + size
                receive(data[idx:nidx])
                idx = nidx
            receive(data[idx:])
            self.assertEqual(self.received, expected)
            del self.received[:]

        # single chunk; i.e. split chunk
        check()
        # individual messages
        check(len(data) // 2)
        # combine from 2 chunks
        check(2, 2, 2)
        # combine from 3 chunks
        check(5, 2)
        # optimal
        check(4, len(msg), 4, len(msg))

        # check message processing closed the protocol
        # once closed, no further data is processed
        proto = self.protocol
        old_receive = proto.receive

        def close(msg):
            proto.close()
            proto.set_receive(old_receive)

        proto.set_receive(close)
        receive(data)
        self.assertEqual(self.received, [])


def to_byte(i):
    return bytes([i])


def _break_mock_cycles(m):
    """break (``mock`` introduced) cycles in mock *m*.

    Do not break cycles due to mock calling.
    """
    m._mock_parent = m._mock_new_parent = None
    for c in m._mock_children.values():
        _break_mock_cycles(c)
    if isinstance(m._mock_return_value, mock.Mock):
        _break_mock_cycles(m._mock_return_value)


class OptimizeTestsBase:
    def setUp(self):
        self.loop = FaithfulLoop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.loop.close()
        asyncio.set_event_loop(None)
        self.assertEqual(self.loop.exceptions, [])


class FutureTestsBase(OptimizeTestsBase):
    def setUp(self):
        super().setUp()
        self.fut = fut = self.make_future(self.loop)
        self.callback = cb = mock.Mock()
        fut.add_done_callback(cb)

    def tearDown(self):
        fut = self.fut
        fut.cancel()
        self.assertEqual(self.loop.exceptions, [])
        _break_mock_cycles(self.callback)
        super().tearDown()

    def test_set_result(self):
        self.fut.set_result(1)
        self.assertEqual(self.fut.result(), 1)
        self.check_called_not_scheduled()

    def test_set_exception(self):
        exc = Exception()
        self.fut.set_exception(exc)
        self.assertIs(self.fut.exception(), exc)
        self.check_called_not_scheduled()

    def test_remove_add_callback(self):
        fut = self.fut
        cb = self.callback
        self.assertEqual(fut.remove_done_callback(cb), 1)
        fut.set_result(1)
        fut.add_done_callback(cb)
        self.check_called_not_scheduled()

    def test_callback_can_add_callback(self):
        fut = self.fut
        cbno = count()
        li = [1, 0, 0, 1]

        def mk_callback():
            """make numbered callback and register it."""
            i = next(cbno)

            @fut.add_done_callback
            def callback(unused):
                li.append(i)
                if not li[i]:
                    # add new callback
                    mk_callback()  # noqa: F821

        mk_callback()
        mk_callback()
        mk_callback()
        fut.set_result(None)
        self.assertEqual(li[4:], [0, 1, 2, 3, 4, 5])
        mk_callback = None  # break reference cycle

    def test_exception_in_callback(self):
        fut = self.fut
        l = []

        def f(_): l.append('f')
        def g(_): l.append('g'); raise RuntimeError('g')
        def h(_): l.append('h')
        def i(_): l.append('i'); raise RuntimeError('i')
        def j(_): l.append('j')

        for x in f,g,h,i,j:
            fut.add_done_callback(x)

        self.assertEqual(self.loop.exceptions, [])
        fut.set_result(None)
        self.assertTrue(fut.done())
        self.assertEqual(l, ['f','g','h','i','j'])

        self.assertEqual(len(self.loop.exceptions), 2)
        e0 = self.loop.exceptions[0]
        e1 = self.loop.exceptions[1]
        self.assertIs(type(e0), dict)
        self.assertIs(type(e1), dict)
        self.assertEqual(set(e0.keys()), {'message', 'exception'})
        self.assertEqual(set(e1.keys()), {'message', 'exception'})
        self.assertRegex(e0['message'], r'Exception in callback .*\bg\b')
        self.assertRegex(e1['message'], r'Exception in callback .*\bi\b')
        # RuntimeError('x') != RuntimeError('x')  -> compare by hand
        x0 = e0['exception']
        x1 = e1['exception']
        self.assertIs(type(x0), RuntimeError)
        self.assertIs(type(x1), RuntimeError)
        self.assertEqual(x0.args, ('g',))
        self.assertEqual(x1.args, ('i',))
        self.loop.exceptions = []

    def test_cancel(self):
        _ = self.fut.cancel('zzz')
        self.assertIs(_, True)
        self.assertTrue(self.fut.cancelled())
        with self.assertRaises(asyncio.CancelledError) as e:
            self.fut.result()
        self.assertEqual(e.exception.args, ('zzz',))
        self.check_called_not_scheduled()

    def check_called_not_scheduled(self):
        fut = self.fut
        self.assertTrue(fut.done())
        self.callback.assert_called_once()
        self.assertEqual(len(self.loop._ready), 0)


class FutureTests(FutureTestsBase, TestCase):
    def make_future(self, loop):
        return Future(loop=self.loop)

class ConcurrentFutureTests(FutureTestsBase, TestCase):
    def make_future(self, loop):
        return ConcurrentFuture(loop=loop)

    def test_result_timeout(self):
        self.assertFalse(self.fut.done())
        with self.assertRaises(asyncio.TimeoutError):
            self.fut.result(0.1)  # should timeout - noone set result on .fut
        self.assertFalse(self.fut.done())


class CoroutineExecutorTestsBase(OptimizeTestsBase):
    def run_loop(self):
        loop = self.loop
        loop.call_soon(loop.stop)
        loop.run_forever()

    def test_noop(self):

        async def noop():
            pass

        t = self.make_task(noop)
        self.assertTrue(t.done())
        self.assertIsNone(t.result())

    def test_async_future(self):
        self.check_future(self.loop.create_future(), True)

    def test_repr(self):

        async def noop():
            pass

        t = self.make_task(noop)
        repr(t)  # satisfied if no exception

    def test_optimized_future(self):
        self.check_future(Future(loop=self.loop))

    def check_future(self, fut, run_loop=False):

        async def wait():
            return await fut

        t = self.make_task(wait)
        self.assertFalse(t.done())
        fut.set_result(1)
        if run_loop:
            self.run_loop()
        self.assertTrue(t.done())
        self.assertEqual(t.result(), 1)

    def test_exception(self):
        fut = Future(loop=self.loop)

        async def exc():
            await fut

        t = self.make_task(exc)
        exc = Exception()
        fut.set_exception(exc)
        self.assertTrue(t.done())
        self.assertIs(t.exception(), exc)
        exc.__traceback__ = None  # break reference cycle

    def test_handled_exception(self):

        fut = Future(loop=self.loop)

        async def exc():
            try:
                await fut
            except Exception:
                return 1

        t = self.make_task(exc)
        exc = Exception()
        fut.set_exception(exc)
        self.assertTrue(t.done())
        self.assertEqual(t.result(), 1)
        exc.__traceback__ = None  # break reference cycle

    def test_cancel_future(self):
        fut = Future(loop=self.loop)

        async def exc():
            await fut

        t = self.make_task(exc)
        _ = fut.cancel('zzz')
        self.assertIs(_, True)
        self.assertTrue(t.done())
        self.assertTrue(t.cancelled())
        with self.assertRaises(asyncio.CancelledError) as e:
            t.result()
        self.assertEqual(e.exception.args, ('zzz',))

    def test_cancel_task_while_blocked_on_optimized_future(self):
        self._test_cancel_task_while_blocked(Future(loop=self.loop))

    def test_cancel_task_while_blocked_on_async_future(self):
        self._test_cancel_task_while_blocked(asyncio.Future(loop=self.loop))

    def _test_cancel_task_while_blocked(self, blocked_on):
        l = []
        go = Future(loop=self.loop)
        waitready = Future(loop=self.loop)
        waiting = blocked_on

        async def f():
            await go  # wait for loop to start
            l.append(1)
            l.append(2)
            waitready.set_result(None)
            await waiting
            l.append(3)

        t = self.make_task(f)
        self.assertFalse(t.done())

        @waitready.add_done_callback
        def _(_):
            _ = t.cancel('zzz')
            self.assertIs(_, True)

        self.loop.call_soon(lambda: go.set_result(None))
        with self.assertRaises(asyncio.CancelledError):
            self.loop.run_until_complete(t)
        self.assertTrue(t.done())
        self.assertTrue(t.cancelled())
        with self.assertRaises(asyncio.CancelledError) as e:
            t.result()
        self.assertEqual(e.exception.args, ('zzz',))
        self.assertTrue(waiting.done())
        self.assertTrue(waiting.cancelled())
        self.assertEqual(l, [1,2])
        _ = t.cancel()
        self.assertIs(_, False)


    def test_cancel_task_while_running(self):
        l = []
        t = None
        go = Future(loop=self.loop)
        waiting = Future(loop=self.loop)

        async def f():
            await go
            l.append(1)
            l.append(2)
            _ = t.cancel('zzz')
            self.assertIs(_, True)
            l.append(3)
            await waiting
            l.append(4)

        t = self.make_task(f)
        self.assertFalse(t.done())
        go.set_result(None)
        self.assertTrue(t.done())
        self.assertTrue(t.cancelled())
        with self.assertRaises(asyncio.CancelledError) as e:
            t.result()
        self.assertEqual(e.exception.args, ('zzz',))
        self.assertTrue(waiting.done())
        self.assertTrue(waiting.cancelled())
        self.assertEqual(l, [1,2,3])


    def test_nested_coro(self):
        l = []

        async def f():
            await g('a')
            await g('b')

        async def g(s):
            await h(s, 1)
            await h(s, 2)
            await h(s, 3)

        async def h(s, n):
            for i in range(1,n+1):
                l.append(s*i)

        t = self.make_task(f)
        self.assertTrue(t.done())
        self.assertEqual(l, ['a', 'a','aa', 'a','aa','aaa', 'b', 'b','bb', 'b','bb','bbb'])

    def test_nest_to_async_coro(self):
        go = Future(loop=self.loop)

        async def f():
            await go  # wait for loop to start before running into asyncio.sleep
            _ = await asyncio.sleep(1, 'zzz')
            return _

        tstart = time()
        t = self.make_task(f)
        self.assertFalse(t.done())
        self.loop.call_soon(lambda: go.set_result(None))
        self.loop.run_until_complete(t)
        tend = time()
        self.assertTrue(t.done())
        self.assertEqual(t.result(), 'zzz')
        self.assertGreaterEqual(tend - tstart, 1)

    def test_return(self):

        # plain return
        async def f():
            return 123

        t = self.make_task(f)
        self.assertTrue(t.done())
        self.assertEqual(t.result(), 123)

        # nested returns
        async def g():
            x = await f()
            return 1000 + x

        t = self.make_task(g)
        self.assertTrue(t.done())
        self.assertEqual(t.result(), 1123)

        # verify that return can go through `except Exception`
        # (but not through `except BaseException`)
        async def h():
            try:
                return 456
                raise Exception('aaa')
            except Exception:
                raise

        t = self.make_task(h)
        self.assertTrue(t.done())
        self.assertEqual(t.result(), 456)


class AsyncTaskTests(CoroutineExecutorTestsBase, TestCase):
    def make_task(self, coro):
        return AsyncTask(coro(), self.loop)


class ConcurrentTaskTests(CoroutineExecutorTestsBase, TestCase):
    def make_task(self, coro):
        loop = self.loop
        t = ConcurrentTask(coro(), loop)
        self.run_loop()
        return t



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Trace file statistics analyzer.

File format:

Each record is 26 bytes, plus a variable number of bytes to store an oid,
with the following layout.  Numbers are big-endian integers.

Offset  Size  Contents

0       4     timestamp (seconds since 1/1/1970)
4       3     data size, in 256-byte increments, rounded up
7       1     code (see below)
8       2     object id length
10      8     start tid
18      8     end tid
26  variable  object id

The code at offset 7 packs three fields:

Mask    bits  Contents

0x80    1     set if there was a non-empty version string
0x7e    6     function and outcome code
0x01    1     current cache file (0 or 1)

The "current cache file" bit is no longer used; it refers to a 2-file
cache scheme used before ZODB 3.3.

The function and outcome codes are documented in detail at the end of
this file in the 'explain' dictionary.  Note that the keys there (and
also the arguments to _trace() in ClientStorage.py) are 'code & 0x7e',
i.e. the low bit is always zero.
"""
import argparse
import gzip
import struct
import sys
import time
# we assign ctime locally to facilitate test replacement!
from time import ctime


def add_interval_argument(parser):
    def _interval(a):
        interval = int(60 * float(a))
        if interval <= 0:
            interval = 60
        elif interval > 3600:
            interval = 3600
        return interval
    parser.add_argument(
        "--interval", "-i",
        default=15*60, type=_interval,
        help="summarizing interval in minutes (default 15; max 60)")


def add_tracefile_argument(parser):

    class GzipFileType(argparse.FileType):
        def __init__(self):
            super().__init__(mode='rb')

        def __call__(self, s):
            f = super().__call__(s)
            if s.endswith(".gz"):
                f = gzip.GzipFile(filename=s, fileobj=f)
            return f

    parser.add_argument("tracefile", type=GzipFileType(),
                        help="The trace to read; may be gzipped")


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    # Parse options
    parser = argparse.ArgumentParser(
                description="Trace file statistics analyzer",
                # Our -h, short for --load-histogram
                # conflicts with default for help, so we handle
                # manually.
                add_help=False)
    verbose_group = parser.add_mutually_exclusive_group()
    verbose_group.add_argument('--verbose', '-v',
                               default=False, action='store_true',
                               help="Be verbose; print each record")
    verbose_group.add_argument('--quiet', '-q',
                               default=False, action='store_true',
                               help="Reduce output; don't print summaries")
    parser.add_argument("--sizes", '-s',
                        default=False, action="store_true",
                        dest="print_size_histogram",
                        help="print histogram of object sizes")
    parser.add_argument("--no-stats", '-S',
                        default=True, action="store_false", dest="dostats",
                        help="don't print statistics")
    parser.add_argument("--load-histogram", "-h",
                        default=False, action="store_true",
                        dest="print_histogram",
                        help="print histogram of object load frequencies")
    parser.add_argument("--check", "-X",
                        default=False, action="store_true", dest="heuristic",
                        help=" enable heuristic checking for misaligned "
                             "records: oids > 2**32"
                             " will be rejected; this requires the tracefile "
                             "to be seekable")
    add_interval_argument(parser)
    add_tracefile_argument(parser)

    if '--help' in args:
        parser.print_help()
        sys.exit(2)

    options = parser.parse_args(args)

    f = options.tracefile

    rt0 = time.time()
    bycode = {}      # map code to count of occurrences
    byinterval = {}  # map code to count in current interval
    records = 0      # number of trace records read
    versions = 0     # number of trace records with versions
    datarecords = 0  # number of records with dlen set
    datasize = 0     # sum of dlen across records with dlen set
    oids = {}        # map oid to number of times it was loaded
    bysize = {}      # map data size to number of loads
    bysizew = {}     # map data size to number of writes
    total_loads = 0
    t0 = None        # first timestamp seen
    te = None        # most recent timestamp seen
    h0 = None        # timestamp at start of current interval
    he = None        # timestamp at end of current interval
    thisinterval = None  # generally te//interval
    f_read = f.read
    unpack = struct.unpack
    FMT = ">iiH8s8s"
    FMT_SIZE = struct.calcsize(FMT)
    assert FMT_SIZE == 26
    # Read file, gathering statistics, and printing each record if verbose.
    print(' '*16, "%7s %7s %7s %7s" % (
            'loads', 'hits', 'inv(h)', 'writes'), end=' ')
    print('hitrate')
    try:
        while 1:
            r = f_read(FMT_SIZE)
            if len(r) < FMT_SIZE:
                break
            ts, code, oidlen, start_tid, end_tid = unpack(FMT, r)
            if ts == 0:
                # Must be a misaligned record caused by a crash.
                if not options.quiet:
                    print("Skipping 8 bytes at offset", f.tell() - FMT_SIZE)
                    f.seek(f.tell() - FMT_SIZE + 8)
                continue
            oid = f_read(oidlen)
            if len(oid) < oidlen:
                break
            records += 1
            if t0 is None:
                t0 = ts
                thisinterval = t0 // options.interval
                h0 = he = ts
            te = ts
            if ts // options.interval != thisinterval:
                if not options.quiet:
                    dumpbyinterval(byinterval, h0, he)
                byinterval = {}
                thisinterval = ts // options.interval
                h0 = ts
            he = ts
            dlen, code = (code & 0x7fffff00) >> 8, code & 0xff
            if dlen:
                datarecords += 1
                datasize += dlen
            if code & 0x80:
                version = 'V'
                versions += 1
            else:
                version = '-'
            code &= 0x7e
            bycode[code] = bycode.get(code, 0) + 1
            byinterval[code] = byinterval.get(code, 0) + 1
            if dlen:
                if code & 0x70 == 0x20:  # All loads
                    bysize[dlen] = d = bysize.get(dlen) or {}
                    d[oid] = d.get(oid, 0) + 1
                elif code & 0x70 == 0x50:  # All stores
                    bysizew[dlen] = d = bysizew.get(dlen) or {}
                    d[oid] = d.get(oid, 0) + 1
            if options.verbose:
                print("%s %02x %s %016x %016x %c%s" % (
                    ctime(ts)[4:-5],
                    code,
                    oid_repr(oid),
                    U64(start_tid),
                    U64(end_tid),
                    version,
                    dlen and (' '+str(dlen)) or ""))
            if code & 0x70 == 0x20:
                oids[oid] = oids.get(oid, 0) + 1
                total_loads += 1
            elif code == 0x00:  # restart
                if not options.quiet:
                    dumpbyinterval(byinterval, h0, he)
                byinterval = {}
                thisinterval = ts // options.interval
                h0 = he = ts
                if not options.quiet:
                    print(ctime(ts)[4:-5], end=' ')
                    print('='*20, "Restart", '='*20)
    except KeyboardInterrupt:
        print("\nInterrupted.  Stats so far:\n")

    end_pos = f.tell()
    f.close()
    rte = time.time()
    if not options.quiet:
        dumpbyinterval(byinterval, h0, he)

    # Error if nothing was read
    if not records:
        print("No records processed", file=sys.stderr)
        return 1

    # Print statistics
    if options.dostats:
        print()
        print("Read {} trace records ({} bytes) in {:.1f} seconds".format(
            addcommas(records), addcommas(end_pos), rte-rt0))
        print("Versions:   %s records used a version" % addcommas(versions))
        print("First time: %s" % ctime(t0))
        print("Last time:  %s" % ctime(te))
        print("Duration:   %s seconds" % addcommas(te-t0))
        print("Data recs:  %s (%.1f%%), average size %d bytes" % (
            addcommas(datarecords),
            100.0 * datarecords / records,
            datasize / datarecords))
        print("Hit rate:   %.1f%% (load hits / loads)" % hitrate(bycode))
        print()
        codes = sorted(bycode.keys())
        print("%13s %4s %s" % ("Count", "Code", "Function (action)"))
        for code in codes:
            print("%13s  %02x  %s" % (
                addcommas(bycode.get(code, 0)),
                code,
                explain.get(code) or "*** unknown code ***"))

    # Print histogram.
    if options.print_histogram:
        print()
        print("Histogram of object load frequency")
        total = len(oids)
        print("Unique oids: %s" % addcommas(total))
        print("Total loads: %s" % addcommas(total_loads))
        s = addcommas(total)
        width = max(len(s), len("objects"))
        fmt = "%5d %" + str(width) + "s %5.1f%% %5.1f%% %5.1f%%"
        hdr = "%5s %" + str(width) + "s %6s %6s %6s"
        print(hdr % ("loads", "objects", "%obj", "%load", "%cum"))
        cum = 0.0
        for binsize, count in histogram(oids):
            obj_percent = 100.0 * count / total
            load_percent = 100.0 * count * binsize / total_loads
            cum += load_percent
            print(fmt % (binsize, addcommas(count),
                         obj_percent, load_percent, cum))

    # Print size histogram.
    if options.print_size_histogram:
        print()
        print("Histograms of object sizes")
        print()
        dumpbysize(bysizew, "written", "writes")
        dumpbysize(bysize, "loaded", "loads")


def dumpbysize(bysize, how, how2):
    print()
    print(f"Unique sizes {how}: {addcommas(len(bysize))}")
    print("%10s %6s %6s" % ("size", "objs", how2))
    sizes = sorted(bysize.keys())
    for size in sizes:
        loads = 0
        for n in bysize[size].values():
            loads += n
        print("%10s %6d %6d" % (addcommas(size),
                                len(bysize.get(size, "")),
                                loads))


def dumpbyinterval(byinterval, h0, he):
    loads = hits = invals = writes = 0
    for code in byinterval:
        if code & 0x20:
            n = byinterval[code]
            loads += n
            if code in (0x22, 0x26):
                hits += n
        elif code & 0x40:
            writes += byinterval[code]
        elif code & 0x10:
            if code != 0x10:
                invals += byinterval[code]

    if loads:
        hr = "%5.1f%%" % (100.0 * hits / loads)
    else:
        hr = 'n/a'

    print("%s-%s %7s %7s %7s %7s %7s" % (
        ctime(h0)[4:-8], ctime(he)[14:-8],
        loads, hits, invals, writes, hr))


def hitrate(bycode):
    loads = hits = 0
    for code in bycode:
        if code & 0x70 == 0x20:
            n = bycode[code]
            loads += n
            if code in (0x22, 0x26):
                hits += n
    if loads:
        return 100.0 * hits / loads
    else:
        return 0.0


def histogram(d):
    bins = {}
    for v in d.values():
        bins[v] = bins.get(v, 0) + 1
    L = sorted(bins.items())
    return L


def U64(s):
    return struct.unpack(">Q", s)[0]


def oid_repr(oid):
    if isinstance(oid, bytes) and len(oid) == 8:
        return '%16x' % U64(oid)
    else:
        return repr(oid)


def addcommas(n):
    sign, s = '', str(n)
    if s[0] == '-':
        sign, s = '-', s[1:]
    i = len(s) - 3
    while i > 0:
        s = s[:i] + ',' + s[i:]
        i -= 3
    return sign + s


explain = {
    # The first hex digit shows the operation, the second the outcome.
    # If the second digit is in "02468" then it is a 'miss'.
    # If it is in "ACE" then it is a 'hit'.

    0x00: "_setup_trace (initialization)",

    0x10: "invalidate (miss)",
    0x1A: "invalidate (hit, version)",
    0x1C: "invalidate (hit, saving non-current)",
    # 0x1E can occur during startup verification.
    0x1E: "invalidate (hit, discarding current or non-current)",

    0x20: "load (miss)",
    0x22: "load (hit)",
    0x24: "load (non-current, miss)",
    0x26: "load (non-current, hit)",

    0x50: "store (version)",
    0x52: "store (current, non-version)",
    0x54: "store (non-current)",
    }

if __name__ == "__main__":
    sys.exit(main())



"""ZEO Protocol.

A ZEO protocol instance can be used as a connection.
It exchanges ``bytes`` messages.
Messages are sent via the methods
``write_message`` (send a single message) and
``write_message_iter`` (send the messages generated by an iterator).
Received messages are reported via callbacks.
Messages are received in the same order as they have been written;
especially, the messages wrote with ``write_message_iter``
are received as contiguous messages.

The first message transmits the protocol version.
Its callback is ``finish_connection``.
The first byte of the protocol version message identifies
an encoding type; the remaining bytes specify the version.
``finish_connection`` is expected to set up
methods ``encode`` and ``decode`` corresponding to the
encoding type.

Followup messages carry encoded tuples
*msgid*, *async_flag*, *name*, *args*
representing either calls (synchronous or asynchronous) or replies.
Their callback is ``message_received``.

ZEO protocol instances can be used concurrently from coroutines (executed
in the same thread).
They are not thread safe.

The ZEO protocol sits on top of a sized message protocol.

The ZEO protocol has client and server variants.
"""
import logging
from asyncio import Protocol

from .smp import SizedMessageProtocol


logger = logging.getLogger(__name__)


class ZEOBaseProtocol(Protocol):
    """ZEO protocol base class for the common features."""

    protocol_version = None

    def __init__(self, loop, name):
        self.loop = loop
        self.name = name

    # API -- defined in ``connection_made``
    # write_message(message)
    # write_message_iter(message_iter)
    def call_async(self, method, args):
        """call method named *method* asynchronously with *args*."""
        self.write_message(self.encode(0, True, method, args))

    def call_async_iter(self, it):
        self.write_message_iter(self.encode(0, True, method, args)
                                for method, args in it)

    def get_peername(self):
        return self.sm_protocol.transport.get_extra_info('peername')

    def protocol_factory(self):
        return self

    closing = None  # ``None`` or closed future
    sm_protocol = None

    def close(self):
        """schedule closing, return closed future."""
        # with ``asyncio``, ``close`` only schedules the closing;
        # close completion is signalled via a call to ``connection_lost``.
        closing = self.closing
        if closing is None:
            closing = self.closing = self.loop.create_future()
            # can get closed before ``sm_protocol`` set up
            if self.sm_protocol is not None:
                # will eventually cause ``connection_lost``
                self.sm_protocol.close()
            else:
                closing.set_result(True)
        elif self.sm_protocol is not None:
            self.sm_protocol.close()  # no problem if repeated
        return closing

    def __repr__(self):
        cls = self.__class__
        return f'{cls.__module__}.{cls.__name__}({self.name})'

    # to be defined by deriving classes
    # def finish_connection(protocol_version_message)
    # def message_received(message)

    #  ``Protocol`` responsibilities -- defined in ``connection_made``
    # data_received
    # eof_received
    # pause_writing
    # resume_writing
    def connection_made(self, transport):
        logger.info("Connected %s", self)
        # set up lower level sized message protocol
        # creates reference cycle
        smp = self.sm_protocol = SizedMessageProtocol(self._first_message)
        smp.connection_made(transport)  # takes over ``transport``
        self.data_received = smp.data_received
        self.eof_received = smp.eof_received
        self.pause_writing = smp.pause_writing
        self.resume_writing = smp.resume_writing
        self.write_message = smp.write_message
        self.write_message_iter = smp.write_message_iter

    # In real life ``connection_lost`` is only called by
    # the transport and ``asyncio.Protocol`` guarantees that
    # it is called exactly once (if ``connection_made`` has
    # been called) or not at all.
    # Some tests, however, call ``connection_lost`` themselves.
    # The following attribute helps to ensure that ``connection_lost``
    # is called exactly once.
    connection_lost_called = False

    def connection_lost(self, exc):
        """The call signals close completion."""
        self.connection_lost_called = True
        self.sm_protocol.connection_lost(exc)
        closing = self.closing
        if closing is None:
            closing = self.closing = self.loop.create_future()
        if not closing.done():
            closing.set_result(True)

    # internal
    def _first_message(self, protocol_version):
        self.sm_protocol.set_receive(self.message_received)
        self.finish_connection(protocol_version)

    # ``uvloop`` workaround
    # We define ``data_received`` in ``connection_made``.
    # ``uvloop``, however, caches ``protocol.data_received`` before
    # it calls ``connection_made`` - at a consequence, data is not
    # received
    # The method below is overridden in ``connection_made``.
    def data_received(self, data):
        self.data_received(data)  # not an infinite loop, because overridden



"""SizedMessageProtocol.

The protocal exchanges ``bytes`` messages.
On the wire, a message is represented by a 4 bytes length followed
by the message's bytes.
"""

import struct
from asyncio import Protocol
from logging import getLogger


logger = getLogger(__name__)


class SizedMessageProtocol(Protocol):
    """asyncio protocol for the exchange of sized messages.

    A protocol object can be used as a connection.
    Write methods are
    ``write_message`` (write a single message)
    and ``write_message_iter`` (write the messages generated by an iterator).
    A received message is reported via the ``receive`` callback.

    Messages are received in the same order as they have been
    written.

    ``SizedMessageProtocol`` instances can be used concurrently
    from coroutines (executed in the same thread).
    They are not thread safe.
    """

    transport = None

    def __init__(self, receive):
        self.receive = receive

    def connection_made(self, transport):
        self.transport = transport

        # output handling
        pack = struct.pack
        paused = False  # if ``paused`` we must buffer output
        output = []  # output buffer - contains messages or iterators
        append = output.append
        writelines = transport.writelines

        def _write_message(message):
            """hand the message over to the transport.

            May set ``paused``.
            """
            writelines((pack(">I", len(message)), message))

        # Note: Outside ``resume_writing``
        # ``not paused`` implies ``not output``. This
        # allows to use ``if paused`` instead of
        # ``if paused or output`` in ``write_message`` and
        # ``write_message_iter``.

        def write_message(message):
            if paused:  # equivalent to ``paused or output``
                append(message)
            else:
                _write_message(message)

        self.write_message = write_message

        def write_message_iter(message_iter):
            it = iter(message_iter)
            if paused:  # equivalent to ``paused or output``
                append(it)
                return
            for message in it:
                _write_message(message)
                if paused:
                    append(it)
                    return

        self.write_message_iter = write_message_iter

        def resume_writing():
            # precondition: ``paused`` and "at least 1 message writable"
            nonlocal paused
            paused = False
            while output and not paused:
                message = output.pop(0)
                if isinstance(message, bytes):
                    # a message
                    _write_message(message)
                else:
                    # an iterator
                    it = message
                    for message in it:
                        _write_message(message)
                        if paused:  # paused again. Put iter back.
                            output.insert(0, it)
                            break
            # post condition: ``paused or not output``

        self.resume_writing = resume_writing

        def pause_writing():
            nonlocal paused
            paused = True

        self.pause_writing = pause_writing

        # input handling
        # the following implements a state machine with
        # states ``process_size``, ``process_message`` and ``closed``
        process_size = 1
        process_message = 2
        closed = None
        read_state = process_size  # current state
        read_wanted = 4  # bytes required for this state
        received_count = 0  # number of received (not yet processed) bytes
        received_buffer = []  # received data chunks
        chunk_index = 0  # first unprocessed byte in first chunk
        unpack = struct.unpack

        def data_received(data):
            nonlocal received_count, read_state, read_wanted, chunk_index
            received_buffer.append(data)
            received_count += len(data)
            # ``not read_state`` means "closed"
            while read_state and read_wanted <= received_count:
                # transfer ``read_wanted`` bytes into ``data``
                data = None
                wanted = read_wanted
                while wanted:
                    chunk = received_buffer[0]
                    ch_unprocessed = len(chunk) - chunk_index
                    if ch_unprocessed > wanted:
                        n_index = chunk_index + wanted
                        fragment = chunk[chunk_index:n_index]
                        chunk_index = n_index
                        wanted = 0
                    else:
                        del received_buffer[0]
                        fragment = \
                            chunk[chunk_index:] if chunk_index else chunk
                        chunk_index = 0
                        wanted -= ch_unprocessed
                    if data is None:  # typical case
                        data = fragment
                    else:
                        data += fragment
                received_count -= read_wanted
                # process ``data``
                if read_state is process_size:
                    read_state = process_message
                    read_wanted = unpack(">I", data)[0]
                else:  # ``read_state is process_message``
                    try:
                        self.receive(data)  # may close: ``not read_state``
                    except Exception:
                        logger.exception("Processing message `%r` failed"
                                         % data)
                    if read_state:  # not yet closed
                        read_state = process_size
                        read_wanted = 4

        # the following introduces a reference cycle broken in ``close``
        self.data_received = data_received

        def eof_received():
            nonlocal read_state
            read_state = closed

        self.eof_received = eof_received

    def set_receive(self, receive):
        self.receive = receive

    __closed = False

    def close(self):
        if self.__closed:
            return
        self.__closed = True
        self.eof_received()
        self.transport.close()
        # break reference cycles
        self.transport = self.receive = self.data_received = None

    # We define ``connection_lost`` to close the transport
    # in order to avoid a ``ResourceWarning``
    # about an unclosed SSL transport -- it should not be necessary
    # as the transport informed us about the lost connection.
    # It also helps for some tests which call ``connection_lost``
    # without transport intervention.
    connection_lost_called = False  # for tests

    def connection_lost(self, exc):
        self.connection_lost_called = True
        if self.__closed:
            return
        self.transport.close()


# Use C implementation if available
try:
    from ._smp import SizedMessageProtocol  # noqa: F811, F401
except ImportError:
    pass



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""ZEO -- Zope Enterprise Objects.

See the file README.txt in this directory for an overview.

ZEO is now part of ZODB; ZODB's home on the web is

    http://wiki.zope.org/ZODB

"""


def client(*args, **kw):
    """
    Shortcut for :class:`ZEO.ClientStorage.ClientStorage`.
    """
    import ZEO.ClientStorage
    return ZEO.ClientStorage.ClientStorage(*args, **kw)


def DB(*args, **kw):
    """
    Shortcut for creating a :class:`ZODB.DB` using a ZEO :func:`~ZEO.client`.
    """
    s = client(*args, **kw)
    try:
        import ZODB
        return ZODB.DB(s)
    except Exception:
        s.close()
        raise


def connection(*args, **kw):
    db = DB(*args, **kw)
    try:
        return db.open_then_close_db_when_connection_closes()
    except Exception:
        db.close()
        raise


def server(path=None, blob_dir=None, storage_conf=None, zeo_conf=None,
           port=0, threaded=True, **kw):
    """Convenience function to start a server for interactive exploration

    This fuction starts a ZEO server, given a storage configuration or
    a file-storage path and blob directory.  You can also supply a ZEO
    configuration string or a port.  If neither a ZEO port or
    configuration is supplied, a port is chosen randomly.

    The server address and a stop function are returned. The address
    can be passed to ZEO.ClientStorage.ClientStorage or ZEO.DB to
    create a client to the server. The stop function can be called
    without arguments to stop the server.

    Arguments:

    path
       A file-storage path. This argument is ignored if a storage
       configuration is supplied.

    blob_dir
       A blob directory path. This argument is ignored if a storage
       configuration is supplied.

    storage_conf
       A storage configuration string.  If none is supplied, then at
       least a file-storage path must be supplied and the storage
       configuration will be generated from the file-storage path and
       the blob directory.

    zeo_conf
       A ZEO server configuration string.

    port
       If no ZEO configuration is supplied, the one will be computed
       from the port.  If no port is supplied, one will be chosedn
       dynamically.

    """
    import ZEO._forker as forker
    if storage_conf is None and path is None:
        storage_conf = '<mappingstorage>\n</mappingstorage>'

    return forker.start_zeo_server(
        storage_conf, zeo_conf, port, keep=True, path=path,
        blob_dir=blob_dir, suicide=False, threaded=threaded, **kw)



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Test setup for ZEO connection logic.

The actual tests are in ConnectionTests.py; this file provides the
platform-dependent scaffolding.
"""


import os

from zope.testing import setupstack

from ZEO.tests import ConnectionTests
from ZEO.tests import InvalidationTests


if os.environ.get('USE_ZOPE_TESTING_DOCTEST'):
    from zope.testing import doctest
else:
    import doctest

import unittest

import ZODB.tests.util

from . import forker


class FileStorageConfig:
    def getConfig(self, path, create, read_only):
        return """\
        <filestorage 1>
        path {}
        create {}
        read-only {}
        </filestorage>""".format(path,
                             create and 'yes' or 'no',
                             read_only and 'yes' or 'no')


class MappingStorageConfig:
    def getConfig(self, path, create, read_only):
        return """<mappingstorage 1/>"""


class FileStorageConnectionTests(
        FileStorageConfig,
        ConnectionTests.ConnectionTests,
        InvalidationTests.InvalidationTests):
    """FileStorage-specific connection tests."""


class FileStorageReconnectionTests(
        FileStorageConfig,
        ConnectionTests.ReconnectionTests):
    """FileStorage-specific re-connection tests."""
    # Run this at level 1 because MappingStorage can't do reconnection tests


class FileStorageInvqTests(
        FileStorageConfig,
        ConnectionTests.InvqTests):
    """FileStorage-specific invalidation queue tests."""


class FileStorageTimeoutTests(
        FileStorageConfig,
        ConnectionTests.TimeoutTests):
    pass


class MappingStorageConnectionTests(
        MappingStorageConfig,
        ConnectionTests.ConnectionTests):
    """Mapping storage connection tests."""

# The ReconnectionTests can't work with MappingStorage because it's only an
# in-memory storage and has no persistent state.


class MappingStorageTimeoutTests(
        MappingStorageConfig,
        ConnectionTests.TimeoutTests):
    pass


class SSLConnectionTests(
        MappingStorageConfig,
        ConnectionTests.SSLConnectionTests):
    pass


test_classes = [FileStorageConnectionTests,
                FileStorageReconnectionTests,
                FileStorageInvqTests,
                FileStorageTimeoutTests,
                MappingStorageConnectionTests,
                MappingStorageTimeoutTests,
                SSLConnectionTests,
                ]


def invalidations_while_connecting():
    r"""
As soon as a client registers with a server, it will recieve
invalidations from the server.  The client must be careful to queue
these invalidations until it is ready to deal with them.  At the time
of the writing of this test, clients weren't careful enough about
queing invalidations.  This led to cache corruption in the form of
both low-level file corruption as well as out-of-date records marked
as current.

This tests tries to provoke this bug by:

- starting a server

    >>> addr, _ = start_server()  # NOQA: F821 undefined name

- opening a client to the server that writes some objects, filling
  it's cache at the same time,

    >>> import ZEO, ZODB.tests.MinPO, transaction
    >>> db = ZEO.DB(addr, client='x')
    >>> conn = db.open()
    >>> nobs = 1000
    >>> for i in range(nobs):
    ...     conn.root()[i] = ZODB.tests.MinPO.MinPO(0)
    >>> transaction.commit()

    >>> import zope.testing.loggingsupport, logging
    >>> handler = zope.testing.loggingsupport.InstalledHandler(
    ...    'ZEO', level=logging.INFO)

    # >>> logging.getLogger('ZEO').debug(
    # ...     'Initial tid %r' % conn.root()._p_serial)

- disconnecting the first client (closing it with a persistent cache),

    >>> db.close()

- starting a second client that writes objects more or less
  constantly; we use ``mods`` and ``allowed_mods`` to avoid
  that the server's invalidation queue is overrun

    >>> import random, threading, time
    >>> stop = False
    >>> allowed_mods = 80  # ``< invalidation_queue_size`` (100)
    >>> mods = 0
    >>> db2 = ZEO.DB(addr)
    >>> tm = transaction.TransactionManager()
    >>> conn2 = db2.open(transaction_manager=tm)
    >>> random = random.Random(0)
    >>> lock = threading.Lock()
    >>> def run():
    ...     global mods
    ...     while 1:
    ...         if stop:
    ...             return
    ...         time.sleep(0)
    ...         with lock:
    ...             if mods >= allowed_mods:
    ...                 continue  # avoid invalidation queue overrun
    ...             mods += 1
    ...             i = random.randint(0, nobs-1)
    ...             conn2.root()[i].value += 1
    ...             tm.commit()
    ...             #logging.getLogger('ZEO').debug(
    ...             #   'COMMIT %s %s %r' % (
    ...             #   i, conn2.root()[i].value, conn2.root()[i]._p_serial))
    >>> thread = threading.Thread(target=run)
    >>> thread.daemon = True
    >>> thread.start()

- restarting the first client, and
- testing for cache validity.

    >>> bad = False
    >>> try:
    ...     for c in range(10):
    ...        time.sleep(.1)
    ...        db = ZODB.DB(ZEO.ClientStorage.ClientStorage(addr, client='x'))
    ...        with lock:
    ...            #logging.getLogger('ZEO').debug('Locked %s' % c)
    ...            msg = "connected and we have caught up"
    ...
    ...            @wait_until(msg, timeout=199)  # NOQA: F821 undefined var
    ...            def _():
    ...                if (db.storage.is_connected()
    ...                        and db.storage.lastTransaction()
    ...                            == db.storage._call('lastTransaction')
    ...                        ):
    ...                    #logging.getLogger('ZEO').debug(
    ...                    #   'Connected %r' % db.storage.lastTransaction())
    ...                    return True
    ...
    ...            conn = db.open()
    ...            for i in range(1000):
    ...                if conn.root()[i].value != conn2.root()[i].value:
    ...                    print('bad', c, i, conn.root()[i].value, end=" ")
    ...                    print(conn2.root()[i].value)
    ...                    bad = True
    ...                    print('client debug log with lock held')
    ...                    while handler.records:
    ...                          record = handler.records.pop(0)
    ...                          print(record.name, record.levelname, end=' ')
    ...                          print(handler.format(record))
    ...            mods = 0  # allow the writer to continue
    ...        #if bad:
    ...        #   with open('server.log') as f:
    ...        #       print(f.read())
    ...        #else:
    ...        #   logging.getLogger('ZEO').debug('GOOD %s' % c)
    ...        db.close()
    ... finally:
    ...     stop = True
    ...     thread.join(10)

    >>> thread.is_alive()
    False

    >>> for record in handler.records:
    ...     if record.levelno < logging.ERROR:
    ...         continue
    ...     print(record.name, record.levelname)
    ...     print(handler.format(record))

    >>> handler.uninstall()

    >>> db.close()
    >>> db2.close()
    """


def test_suite():
    suite = unittest.TestSuite()

    for klass in test_classes:
        test_loader = unittest.TestLoader()
        test_loader.testMethodPrefix = 'check'
        sub = test_loader.loadTestsFromTestCase(klass)
        sub.layer = ZODB.tests.util.MininalTestLayer(
            klass.__name__ + ' ZEO Connection Tests')
        suite.addTest(sub)

    sub = doctest.DocTestSuite(
        setUp=forker.setUp, tearDown=setupstack.tearDown,
        )
    sub.layer = ZODB.tests.util.MininalTestLayer('ZEO Connection DocTests')
    suite.addTest(sub)

    return suite



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""A ZEO client-server stress test to look for leaks.

The stress test should run in an infinite loop and should involve
multiple connections.
"""
# TODO: This code is currently broken.

import os
import random

import transaction
import ZODB
from ZODB.MappingStorage import MappingStorage
from ZODB.tests import MinPO

from ZEO.ClientStorage import ClientStorage
from ZEO.tests import forker


NUM_TRANSACTIONS_PER_CONN = 10
NUM_CONNECTIONS = 10
NUM_ROOTS = 20
MAX_DEPTH = 20
MIN_OBJSIZE = 128
MAX_OBJSIZE = 2048


def an_object():
    """Return an object suitable for a PersistentMapping key"""
    size = random.randrange(MIN_OBJSIZE, MAX_OBJSIZE)
    if os.path.exists("/dev/urandom"):
        fp = open("/dev/urandom")
        buf = fp.read(size)
        fp.close()
        return buf
    else:
        fp = open(MinPO.__file__)
        lst = list(fp.read(size))
        fp.close()
        random.shuffle(lst)
        return "".join(lst)


def setup(cn):
    """Initialize the database with some objects"""
    root = cn.root()
    for i in range(NUM_ROOTS):
        prev = an_object()
        for j in range(random.randrange(1, MAX_DEPTH)):
            o = MinPO.MinPO(prev)
            prev = o
        root[an_object()] = o
        transaction.commit()
    cn.close()


def work(cn):
    """Do some work with a transaction"""
    cn.sync()
    root = cn.root()
    obj = random.choice(root.values())
    # walk down to the bottom
    while not isinstance(obj.value, str):
        obj = obj.value
    obj.value = an_object()
    transaction.commit()


def main():
    # Yuck!  Need to cleanup forker so that the API is consistent
    # across Unix and Windows, at least if that's possible.
    if os.name == "nt":
        zaddr, tport, pid = forker.start_zeo_server('MappingStorage', ())

        def exitserver():
            import socket
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(tport)
            s.close()
    else:
        zaddr = '', random.randrange(20000, 30000)
        pid, exitobj = forker.start_zeo_server(MappingStorage(), zaddr)

        def exitserver():
            exitobj.close()

    while 1:
        pid = start_child(zaddr)
        print("started", pid)
        os.waitpid(pid, 0)

    exitserver()


def start_child(zaddr):

    pid = os.fork()
    if pid != 0:
        return pid
    try:
        _start_child(zaddr)
    finally:
        os._exit(0)


def _start_child(zaddr):
    storage = ClientStorage(zaddr, debug=1, min_disconnect_poll=0.5, wait=1)
    db = ZODB.DB(storage, pool_size=NUM_CONNECTIONS)
    setup(db.open())
    conns = []
    conn_count = 0

    for i in range(NUM_CONNECTIONS):
        c = db.open()
        c.__count = 0
        conns.append(c)
        conn_count += 1

    while conn_count < 25:
        c = random.choice(conns)
        if c.__count > NUM_TRANSACTIONS_PER_CONN:
            conns.remove(c)
            c.close()
            conn_count += 1
            c = db.open()
            c.__count = 0
            conns.append(c)
        else:
            c.__count += 1
        work(c)


if __name__ == "__main__":
    main()
